import datetime
import logging
import os

import numpy as np
import yaml
from lightning.pytorch import seed_everything

from fastprop import fastprop_core
from fastprop.defaults import _init_loggers, init_logger
from fastprop.preprocessing import preprocess

# choose the descriptor set absed on the args
from fastprop.utils import _get_descs, load_from_csv

logger = init_logger(__name__)


def train_fastprop(
    output_directory,
    input_file,
    smiles_column,
    target_columns,
    descriptors="optimized",
    enable_cache=True,
    precomputed=None,
    zero_variance_drop=False,
    colinear_drop=False,
    fnn_layers=2,
    hidden_size=1800,
    learning_rate=0.0001,
    batch_size=2048,
    number_epochs=100,
    number_repeats=1,
    problem_type="regression",
    checkpoint=None,
    train_size=0.8,
    val_size=0.1,
    test_size=0.1,
    sampler="random",
    random_seed=0,
    patience=5,
):
    """Driver function for automatic training.

    See the fastprop documentation or CLI --help for details on each argument.
    """
    if checkpoint is not None:
        raise RuntimeError("TODO: Restarting from checkpoint not currently supported. Exiting.")
    # make output directories
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)
    output_subdirectory = os.path.join(output_directory, f"fastprop_{int(datetime.datetime.utcnow().timestamp())}")
    os.mkdir(output_subdirectory)
    os.mkdir(os.path.join(output_subdirectory, "checkpoints"))
    _init_loggers(output_subdirectory)
    logging.getLogger("pytorch_lightning").setLevel(logging.INFO)
    seed_everything(random_seed)
    targets, mols, smiles = load_from_csv(input_file, smiles_column, target_columns)
    descs = _get_descs(precomputed, input_file, output_directory, descriptors, enable_cache, mols, as_df=True)

    logger.info("Preprocessing features")
    X = preprocess(descs, zero_variance_drop, colinear_drop)

    # write information needed for feature importance, prediction, etc. into the checkpoints directory for later use
    with open(os.path.join(output_subdirectory, "checkpoints", "fastprop_config.yml"), "w") as file:
        file.write(
            yaml.dump(
                dict(smiles=smiles_column, targets=target_columns, descriptors=X.columns.to_list()),
                sort_keys=False,
            )
        )

    input_size = X.shape[1]
    readout_size = targets.shape[1] if problem_type != "multiclass" else int(np.max(targets) + 1)

    return fastprop_core._training_loop(
        number_repeats,
        number_epochs,
        input_size,
        hidden_size,
        readout_size,
        learning_rate,
        fnn_layers,
        output_subdirectory,
        patience,
        problem_type,
        train_size,
        val_size,
        test_size,
        sampler,
        smiles,
        X.to_numpy(),
        targets,
        target_columns,
        batch_size,
        random_seed,
        hopt=False,
    )



from .calculate_descriptors import _get_descs, calculate_mordred_desciptors
from .descriptor_lists import ALL_2D, SUBSET_947, descriptors_lookup
from .load_data import load_from_csv, load_saved_desc
from .select_descriptors import mordred_descriptors_from_strings
from .validate_config import validate_config

__all__ = [
    "_get_descs",
    "calculate_mordred_desciptors",
    "ALL_2D",
    "SUBSET_947",
    "descriptors_lookup",
    "load_from_csv",
    "load_saved_desc",
    "mordred_descriptors_from_strings",
    "validate_config",
]



# Complete list of 2D descriptors implemented in Mordred as well as a subset
# of those descriptors which are often useful for fastprop after preprocessing

ALL_2D = [
    "ABC",
    "ABCGG",
    "nAcid",
    "nBase",
    "SpAbs_A",
    "SpMax_A",
    "SpDiam_A",
    "SpAD_A",
    "SpMAD_A",
    "LogEE_A",
    "VE1_A",
    "VE2_A",
    "VE3_A",
    "VR1_A",
    "VR2_A",
    "VR3_A",
    "nAromAtom",
    "nAromBond",
    "nAtom",
    "nHeavyAtom",
    "nSpiro",
    "nBridgehead",
    "nHetero",
    "nH",
    "nB",
    "nC",
    "nN",
    "nO",
    "nS",
    "nP",
    "nF",
    "nCl",
    "nBr",
    "nI",
    "nX",
    "ATS0dv",
    "ATS1dv",
    "ATS2dv",
    "ATS3dv",
    "ATS4dv",
    "ATS5dv",
    "ATS6dv",
    "ATS7dv",
    "ATS8dv",
    "ATS0d",
    "ATS1d",
    "ATS2d",
    "ATS3d",
    "ATS4d",
    "ATS5d",
    "ATS6d",
    "ATS7d",
    "ATS8d",
    "ATS0s",
    "ATS1s",
    "ATS2s",
    "ATS3s",
    "ATS4s",
    "ATS5s",
    "ATS6s",
    "ATS7s",
    "ATS8s",
    "ATS0Z",
    "ATS1Z",
    "ATS2Z",
    "ATS3Z",
    "ATS4Z",
    "ATS5Z",
    "ATS6Z",
    "ATS7Z",
    "ATS8Z",
    "ATS0m",
    "ATS1m",
    "ATS2m",
    "ATS3m",
    "ATS4m",
    "ATS5m",
    "ATS6m",
    "ATS7m",
    "ATS8m",
    "ATS0v",
    "ATS1v",
    "ATS2v",
    "ATS3v",
    "ATS4v",
    "ATS5v",
    "ATS6v",
    "ATS7v",
    "ATS8v",
    "ATS0se",
    "ATS1se",
    "ATS2se",
    "ATS3se",
    "ATS4se",
    "ATS5se",
    "ATS6se",
    "ATS7se",
    "ATS8se",
    "ATS0pe",
    "ATS1pe",
    "ATS2pe",
    "ATS3pe",
    "ATS4pe",
    "ATS5pe",
    "ATS6pe",
    "ATS7pe",
    "ATS8pe",
    "ATS0are",
    "ATS1are",
    "ATS2are",
    "ATS3are",
    "ATS4are",
    "ATS5are",
    "ATS6are",
    "ATS7are",
    "ATS8are",
    "ATS0p",
    "ATS1p",
    "ATS2p",
    "ATS3p",
    "ATS4p",
    "ATS5p",
    "ATS6p",
    "ATS7p",
    "ATS8p",
    "ATS0i",
    "ATS1i",
    "ATS2i",
    "ATS3i",
    "ATS4i",
    "ATS5i",
    "ATS6i",
    "ATS7i",
    "ATS8i",
    "AATS0dv",
    "AATS1dv",
    "AATS2dv",
    "AATS3dv",
    "AATS4dv",
    "AATS5dv",
    "AATS6dv",
    "AATS7dv",
    "AATS8dv",
    "AATS0d",
    "AATS1d",
    "AATS2d",
    "AATS3d",
    "AATS4d",
    "AATS5d",
    "AATS6d",
    "AATS7d",
    "AATS8d",
    "AATS0s",
    "AATS1s",
    "AATS2s",
    "AATS3s",
    "AATS4s",
    "AATS5s",
    "AATS6s",
    "AATS7s",
    "AATS8s",
    "AATS0Z",
    "AATS1Z",
    "AATS2Z",
    "AATS3Z",
    "AATS4Z",
    "AATS5Z",
    "AATS6Z",
    "AATS7Z",
    "AATS8Z",
    "AATS0m",
    "AATS1m",
    "AATS2m",
    "AATS3m",
    "AATS4m",
    "AATS5m",
    "AATS6m",
    "AATS7m",
    "AATS8m",
    "AATS0v",
    "AATS1v",
    "AATS2v",
    "AATS3v",
    "AATS4v",
    "AATS5v",
    "AATS6v",
    "AATS7v",
    "AATS8v",
    "AATS0se",
    "AATS1se",
    "AATS2se",
    "AATS3se",
    "AATS4se",
    "AATS5se",
    "AATS6se",
    "AATS7se",
    "AATS8se",
    "AATS0pe",
    "AATS1pe",
    "AATS2pe",
    "AATS3pe",
    "AATS4pe",
    "AATS5pe",
    "AATS6pe",
    "AATS7pe",
    "AATS8pe",
    "AATS0are",
    "AATS1are",
    "AATS2are",
    "AATS3are",
    "AATS4are",
    "AATS5are",
    "AATS6are",
    "AATS7are",
    "AATS8are",
    "AATS0p",
    "AATS1p",
    "AATS2p",
    "AATS3p",
    "AATS4p",
    "AATS5p",
    "AATS6p",
    "AATS7p",
    "AATS8p",
    "AATS0i",
    "AATS1i",
    "AATS2i",
    "AATS3i",
    "AATS4i",
    "AATS5i",
    "AATS6i",
    "AATS7i",
    "AATS8i",
    "ATSC0c",
    "ATSC1c",
    "ATSC2c",
    "ATSC3c",
    "ATSC4c",
    "ATSC5c",
    "ATSC6c",
    "ATSC7c",
    "ATSC8c",
    "ATSC0dv",
    "ATSC1dv",
    "ATSC2dv",
    "ATSC3dv",
    "ATSC4dv",
    "ATSC5dv",
    "ATSC6dv",
    "ATSC7dv",
    "ATSC8dv",
    "ATSC0d",
    "ATSC1d",
    "ATSC2d",
    "ATSC3d",
    "ATSC4d",
    "ATSC5d",
    "ATSC6d",
    "ATSC7d",
    "ATSC8d",
    "ATSC0s",
    "ATSC1s",
    "ATSC2s",
    "ATSC3s",
    "ATSC4s",
    "ATSC5s",
    "ATSC6s",
    "ATSC7s",
    "ATSC8s",
    "ATSC0Z",
    "ATSC1Z",
    "ATSC2Z",
    "ATSC3Z",
    "ATSC4Z",
    "ATSC5Z",
    "ATSC6Z",
    "ATSC7Z",
    "ATSC8Z",
    "ATSC0m",
    "ATSC1m",
    "ATSC2m",
    "ATSC3m",
    "ATSC4m",
    "ATSC5m",
    "ATSC6m",
    "ATSC7m",
    "ATSC8m",
    "ATSC0v",
    "ATSC1v",
    "ATSC2v",
    "ATSC3v",
    "ATSC4v",
    "ATSC5v",
    "ATSC6v",
    "ATSC7v",
    "ATSC8v",
    "ATSC0se",
    "ATSC1se",
    "ATSC2se",
    "ATSC3se",
    "ATSC4se",
    "ATSC5se",
    "ATSC6se",
    "ATSC7se",
    "ATSC8se",
    "ATSC0pe",
    "ATSC1pe",
    "ATSC2pe",
    "ATSC3pe",
    "ATSC4pe",
    "ATSC5pe",
    "ATSC6pe",
    "ATSC7pe",
    "ATSC8pe",
    "ATSC0are",
    "ATSC1are",
    "ATSC2are",
    "ATSC3are",
    "ATSC4are",
    "ATSC5are",
    "ATSC6are",
    "ATSC7are",
    "ATSC8are",
    "ATSC0p",
    "ATSC1p",
    "ATSC2p",
    "ATSC3p",
    "ATSC4p",
    "ATSC5p",
    "ATSC6p",
    "ATSC7p",
    "ATSC8p",
    "ATSC0i",
    "ATSC1i",
    "ATSC2i",
    "ATSC3i",
    "ATSC4i",
    "ATSC5i",
    "ATSC6i",
    "ATSC7i",
    "ATSC8i",
    "AATSC0c",
    "AATSC1c",
    "AATSC2c",
    "AATSC3c",
    "AATSC4c",
    "AATSC5c",
    "AATSC6c",
    "AATSC7c",
    "AATSC8c",
    "AATSC0dv",
    "AATSC1dv",
    "AATSC2dv",
    "AATSC3dv",
    "AATSC4dv",
    "AATSC5dv",
    "AATSC6dv",
    "AATSC7dv",
    "AATSC8dv",
    "AATSC0d",
    "AATSC1d",
    "AATSC2d",
    "AATSC3d",
    "AATSC4d",
    "AATSC5d",
    "AATSC6d",
    "AATSC7d",
    "AATSC8d",
    "AATSC0s",
    "AATSC1s",
    "AATSC2s",
    "AATSC3s",
    "AATSC4s",
    "AATSC5s",
    "AATSC6s",
    "AATSC7s",
    "AATSC8s",
    "AATSC0Z",
    "AATSC1Z",
    "AATSC2Z",
    "AATSC3Z",
    "AATSC4Z",
    "AATSC5Z",
    "AATSC6Z",
    "AATSC7Z",
    "AATSC8Z",
    "AATSC0m",
    "AATSC1m",
    "AATSC2m",
    "AATSC3m",
    "AATSC4m",
    "AATSC5m",
    "AATSC6m",
    "AATSC7m",
    "AATSC8m",
    "AATSC0v",
    "AATSC1v",
    "AATSC2v",
    "AATSC3v",
    "AATSC4v",
    "AATSC5v",
    "AATSC6v",
    "AATSC7v",
    "AATSC8v",
    "AATSC0se",
    "AATSC1se",
    "AATSC2se",
    "AATSC3se",
    "AATSC4se",
    "AATSC5se",
    "AATSC6se",
    "AATSC7se",
    "AATSC8se",
    "AATSC0pe",
    "AATSC1pe",
    "AATSC2pe",
    "AATSC3pe",
    "AATSC4pe",
    "AATSC5pe",
    "AATSC6pe",
    "AATSC7pe",
    "AATSC8pe",
    "AATSC0are",
    "AATSC1are",
    "AATSC2are",
    "AATSC3are",
    "AATSC4are",
    "AATSC5are",
    "AATSC6are",
    "AATSC7are",
    "AATSC8are",
    "AATSC0p",
    "AATSC1p",
    "AATSC2p",
    "AATSC3p",
    "AATSC4p",
    "AATSC5p",
    "AATSC6p",
    "AATSC7p",
    "AATSC8p",
    "AATSC0i",
    "AATSC1i",
    "AATSC2i",
    "AATSC3i",
    "AATSC4i",
    "AATSC5i",
    "AATSC6i",
    "AATSC7i",
    "AATSC8i",
    "MATS1c",
    "MATS2c",
    "MATS3c",
    "MATS4c",
    "MATS5c",
    "MATS6c",
    "MATS7c",
    "MATS8c",
    "MATS1dv",
    "MATS2dv",
    "MATS3dv",
    "MATS4dv",
    "MATS5dv",
    "MATS6dv",
    "MATS7dv",
    "MATS8dv",
    "MATS1d",
    "MATS2d",
    "MATS3d",
    "MATS4d",
    "MATS5d",
    "MATS6d",
    "MATS7d",
    "MATS8d",
    "MATS1s",
    "MATS2s",
    "MATS3s",
    "MATS4s",
    "MATS5s",
    "MATS6s",
    "MATS7s",
    "MATS8s",
    "MATS1Z",
    "MATS2Z",
    "MATS3Z",
    "MATS4Z",
    "MATS5Z",
    "MATS6Z",
    "MATS7Z",
    "MATS8Z",
    "MATS1m",
    "MATS2m",
    "MATS3m",
    "MATS4m",
    "MATS5m",
    "MATS6m",
    "MATS7m",
    "MATS8m",
    "MATS1v",
    "MATS2v",
    "MATS3v",
    "MATS4v",
    "MATS5v",
    "MATS6v",
    "MATS7v",
    "MATS8v",
    "MATS1se",
    "MATS2se",
    "MATS3se",
    "MATS4se",
    "MATS5se",
    "MATS6se",
    "MATS7se",
    "MATS8se",
    "MATS1pe",
    "MATS2pe",
    "MATS3pe",
    "MATS4pe",
    "MATS5pe",
    "MATS6pe",
    "MATS7pe",
    "MATS8pe",
    "MATS1are",
    "MATS2are",
    "MATS3are",
    "MATS4are",
    "MATS5are",
    "MATS6are",
    "MATS7are",
    "MATS8are",
    "MATS1p",
    "MATS2p",
    "MATS3p",
    "MATS4p",
    "MATS5p",
    "MATS6p",
    "MATS7p",
    "MATS8p",
    "MATS1i",
    "MATS2i",
    "MATS3i",
    "MATS4i",
    "MATS5i",
    "MATS6i",
    "MATS7i",
    "MATS8i",
    "GATS1c",
    "GATS2c",
    "GATS3c",
    "GATS4c",
    "GATS5c",
    "GATS6c",
    "GATS7c",
    "GATS8c",
    "GATS1dv",
    "GATS2dv",
    "GATS3dv",
    "GATS4dv",
    "GATS5dv",
    "GATS6dv",
    "GATS7dv",
    "GATS8dv",
    "GATS1d",
    "GATS2d",
    "GATS3d",
    "GATS4d",
    "GATS5d",
    "GATS6d",
    "GATS7d",
    "GATS8d",
    "GATS1s",
    "GATS2s",
    "GATS3s",
    "GATS4s",
    "GATS5s",
    "GATS6s",
    "GATS7s",
    "GATS8s",
    "GATS1Z",
    "GATS2Z",
    "GATS3Z",
    "GATS4Z",
    "GATS5Z",
    "GATS6Z",
    "GATS7Z",
    "GATS8Z",
    "GATS1m",
    "GATS2m",
    "GATS3m",
    "GATS4m",
    "GATS5m",
    "GATS6m",
    "GATS7m",
    "GATS8m",
    "GATS1v",
    "GATS2v",
    "GATS3v",
    "GATS4v",
    "GATS5v",
    "GATS6v",
    "GATS7v",
    "GATS8v",
    "GATS1se",
    "GATS2se",
    "GATS3se",
    "GATS4se",
    "GATS5se",
    "GATS6se",
    "GATS7se",
    "GATS8se",
    "GATS1pe",
    "GATS2pe",
    "GATS3pe",
    "GATS4pe",
    "GATS5pe",
    "GATS6pe",
    "GATS7pe",
    "GATS8pe",
    "GATS1are",
    "GATS2are",
    "GATS3are",
    "GATS4are",
    "GATS5are",
    "GATS6are",
    "GATS7are",
    "GATS8are",
    "GATS1p",
    "GATS2p",
    "GATS3p",
    "GATS4p",
    "GATS5p",
    "GATS6p",
    "GATS7p",
    "GATS8p",
    "GATS1i",
    "GATS2i",
    "GATS3i",
    "GATS4i",
    "GATS5i",
    "GATS6i",
    "GATS7i",
    "GATS8i",
    "BCUTc-1h",
    "BCUTc-1l",
    "BCUTdv-1h",
    "BCUTdv-1l",
    "BCUTd-1h",
    "BCUTd-1l",
    "BCUTs-1h",
    "BCUTs-1l",
    "BCUTZ-1h",
    "BCUTZ-1l",
    "BCUTm-1h",
    "BCUTm-1l",
    "BCUTv-1h",
    "BCUTv-1l",
    "BCUTse-1h",
    "BCUTse-1l",
    "BCUTpe-1h",
    "BCUTpe-1l",
    "BCUTare-1h",
    "BCUTare-1l",
    "BCUTp-1h",
    "BCUTp-1l",
    "BCUTi-1h",
    "BCUTi-1l",
    "BalabanJ",
    "SpAbs_DzZ",
    "SpMax_DzZ",
    "SpDiam_DzZ",
    "SpAD_DzZ",
    "SpMAD_DzZ",
    "LogEE_DzZ",
    "SM1_DzZ",
    "VE1_DzZ",
    "VE2_DzZ",
    "VE3_DzZ",
    "VR1_DzZ",
    "VR2_DzZ",
    "VR3_DzZ",
    "SpAbs_Dzm",
    "SpMax_Dzm",
    "SpDiam_Dzm",
    "SpAD_Dzm",
    "SpMAD_Dzm",
    "LogEE_Dzm",
    "SM1_Dzm",
    "VE1_Dzm",
    "VE2_Dzm",
    "VE3_Dzm",
    "VR1_Dzm",
    "VR2_Dzm",
    "VR3_Dzm",
    "SpAbs_Dzv",
    "SpMax_Dzv",
    "SpDiam_Dzv",
    "SpAD_Dzv",
    "SpMAD_Dzv",
    "LogEE_Dzv",
    "SM1_Dzv",
    "VE1_Dzv",
    "VE2_Dzv",
    "VE3_Dzv",
    "VR1_Dzv",
    "VR2_Dzv",
    "VR3_Dzv",
    "SpAbs_Dzse",
    "SpMax_Dzse",
    "SpDiam_Dzse",
    "SpAD_Dzse",
    "SpMAD_Dzse",
    "LogEE_Dzse",
    "SM1_Dzse",
    "VE1_Dzse",
    "VE2_Dzse",
    "VE3_Dzse",
    "VR1_Dzse",
    "VR2_Dzse",
    "VR3_Dzse",
    "SpAbs_Dzpe",
    "SpMax_Dzpe",
    "SpDiam_Dzpe",
    "SpAD_Dzpe",
    "SpMAD_Dzpe",
    "LogEE_Dzpe",
    "SM1_Dzpe",
    "VE1_Dzpe",
    "VE2_Dzpe",
    "VE3_Dzpe",
    "VR1_Dzpe",
    "VR2_Dzpe",
    "VR3_Dzpe",
    "SpAbs_Dzare",
    "SpMax_Dzare",
    "SpDiam_Dzare",
    "SpAD_Dzare",
    "SpMAD_Dzare",
    "LogEE_Dzare",
    "SM1_Dzare",
    "VE1_Dzare",
    "VE2_Dzare",
    "VE3_Dzare",
    "VR1_Dzare",
    "VR2_Dzare",
    "VR3_Dzare",
    "SpAbs_Dzp",
    "SpMax_Dzp",
    "SpDiam_Dzp",
    "SpAD_Dzp",
    "SpMAD_Dzp",
    "LogEE_Dzp",
    "SM1_Dzp",
    "VE1_Dzp",
    "VE2_Dzp",
    "VE3_Dzp",
    "VR1_Dzp",
    "VR2_Dzp",
    "VR3_Dzp",
    "SpAbs_Dzi",
    "SpMax_Dzi",
    "SpDiam_Dzi",
    "SpAD_Dzi",
    "SpMAD_Dzi",
    "LogEE_Dzi",
    "SM1_Dzi",
    "VE1_Dzi",
    "VE2_Dzi",
    "VE3_Dzi",
    "VR1_Dzi",
    "VR2_Dzi",
    "VR3_Dzi",
    "BertzCT",
    "nBonds",
    "nBondsO",
    "nBondsS",
    "nBondsD",
    "nBondsT",
    "nBondsA",
    "nBondsM",
    "nBondsKS",
    "nBondsKD",
    "RNCG",
    "RPCG",
    "C1SP1",
    "C2SP1",
    "C1SP2",
    "C2SP2",
    "C3SP2",
    "C1SP3",
    "C2SP3",
    "C3SP3",
    "C4SP3",
    "HybRatio",
    "FCSP3",
    "Xch-3d",
    "Xch-4d",
    "Xch-5d",
    "Xch-6d",
    "Xch-7d",
    "Xch-3dv",
    "Xch-4dv",
    "Xch-5dv",
    "Xch-6dv",
    "Xch-7dv",
    "Xc-3d",
    "Xc-4d",
    "Xc-5d",
    "Xc-6d",
    "Xc-3dv",
    "Xc-4dv",
    "Xc-5dv",
    "Xc-6dv",
    "Xpc-4d",
    "Xpc-5d",
    "Xpc-6d",
    "Xpc-4dv",
    "Xpc-5dv",
    "Xpc-6dv",
    "Xp-0d",
    "Xp-1d",
    "Xp-2d",
    "Xp-3d",
    "Xp-4d",
    "Xp-5d",
    "Xp-6d",
    "Xp-7d",
    "AXp-0d",
    "AXp-1d",
    "AXp-2d",
    "AXp-3d",
    "AXp-4d",
    "AXp-5d",
    "AXp-6d",
    "AXp-7d",
    "Xp-0dv",
    "Xp-1dv",
    "Xp-2dv",
    "Xp-3dv",
    "Xp-4dv",
    "Xp-5dv",
    "Xp-6dv",
    "Xp-7dv",
    "AXp-0dv",
    "AXp-1dv",
    "AXp-2dv",
    "AXp-3dv",
    "AXp-4dv",
    "AXp-5dv",
    "AXp-6dv",
    "AXp-7dv",
    "SZ",
    "Sm",
    "Sv",
    "Sse",
    "Spe",
    "Sare",
    "Sp",
    "Si",
    "MZ",
    "Mm",
    "Mv",
    "Mse",
    "Mpe",
    "Mare",
    "Mp",
    "Mi",
    "SpAbs_Dt",
    "SpMax_Dt",
    "SpDiam_Dt",
    "SpAD_Dt",
    "SpMAD_Dt",
    "LogEE_Dt",
    "SM1_Dt",
    "VE1_Dt",
    "VE2_Dt",
    "VE3_Dt",
    "VR1_Dt",
    "VR2_Dt",
    "VR3_Dt",
    "DetourIndex",
    "SpAbs_D",
    "SpMax_D",
    "SpDiam_D",
    "SpAD_D",
    "SpMAD_D",
    "LogEE_D",
    "VE1_D",
    "VE2_D",
    "VE3_D",
    "VR1_D",
    "VR2_D",
    "VR3_D",
    "NsLi",
    "NssBe",
    "NssssBe",
    "NssBH",
    "NsssB",
    "NssssB",
    "NsCH3",
    "NdCH2",
    "NssCH2",
    "NtCH",
    "NdsCH",
    "NaaCH",
    "NsssCH",
    "NddC",
    "NtsC",
    "NdssC",
    "NaasC",
    "NaaaC",
    "NssssC",
    "NsNH3",
    "NsNH2",
    "NssNH2",
    "NdNH",
    "NssNH",
    "NaaNH",
    "NtN",
    "NsssNH",
    "NdsN",
    "NaaN",
    "NsssN",
    "NddsN",
    "NaasN",
    "NssssN",
    "NsOH",
    "NdO",
    "NssO",
    "NaaO",
    "NsF",
    "NsSiH3",
    "NssSiH2",
    "NsssSiH",
    "NssssSi",
    "NsPH2",
    "NssPH",
    "NsssP",
    "NdsssP",
    "NsssssP",
    "NsSH",
    "NdS",
    "NssS",
    "NaaS",
    "NdssS",
    "NddssS",
    "NsCl",
    "NsGeH3",
    "NssGeH2",
    "NsssGeH",
    "NssssGe",
    "NsAsH2",
    "NssAsH",
    "NsssAs",
    "NsssdAs",
    "NsssssAs",
    "NsSeH",
    "NdSe",
    "NssSe",
    "NaaSe",
    "NdssSe",
    "NddssSe",
    "NsBr",
    "NsSnH3",
    "NssSnH2",
    "NsssSnH",
    "NssssSn",
    "NsI",
    "NsPbH3",
    "NssPbH2",
    "NsssPbH",
    "NssssPb",
    "SsLi",
    "SssBe",
    "SssssBe",
    "SssBH",
    "SsssB",
    "SssssB",
    "SsCH3",
    "SdCH2",
    "SssCH2",
    "StCH",
    "SdsCH",
    "SaaCH",
    "SsssCH",
    "SddC",
    "StsC",
    "SdssC",
    "SaasC",
    "SaaaC",
    "SssssC",
    "SsNH3",
    "SsNH2",
    "SssNH2",
    "SdNH",
    "SssNH",
    "SaaNH",
    "StN",
    "SsssNH",
    "SdsN",
    "SaaN",
    "SsssN",
    "SddsN",
    "SaasN",
    "SssssN",
    "SsOH",
    "SdO",
    "SssO",
    "SaaO",
    "SsF",
    "SsSiH3",
    "SssSiH2",
    "SsssSiH",
    "SssssSi",
    "SsPH2",
    "SssPH",
    "SsssP",
    "SdsssP",
    "SsssssP",
    "SsSH",
    "SdS",
    "SssS",
    "SaaS",
    "SdssS",
    "SddssS",
    "SsCl",
    "SsGeH3",
    "SssGeH2",
    "SsssGeH",
    "SssssGe",
    "SsAsH2",
    "SssAsH",
    "SsssAs",
    "SsssdAs",
    "SsssssAs",
    "SsSeH",
    "SdSe",
    "SssSe",
    "SaaSe",
    "SdssSe",
    "SddssSe",
    "SsBr",
    "SsSnH3",
    "SssSnH2",
    "SsssSnH",
    "SssssSn",
    "SsI",
    "SsPbH3",
    "SssPbH2",
    "SsssPbH",
    "SssssPb",
    "MAXsLi",
    "MAXssBe",
    "MAXssssBe",
    "MAXssBH",
    "MAXsssB",
    "MAXssssB",
    "MAXsCH3",
    "MAXdCH2",
    "MAXssCH2",
    "MAXtCH",
    "MAXdsCH",
    "MAXaaCH",
    "MAXsssCH",
    "MAXddC",
    "MAXtsC",
    "MAXdssC",
    "MAXaasC",
    "MAXaaaC",
    "MAXssssC",
    "MAXsNH3",
    "MAXsNH2",
    "MAXssNH2",
    "MAXdNH",
    "MAXssNH",
    "MAXaaNH",
    "MAXtN",
    "MAXsssNH",
    "MAXdsN",
    "MAXaaN",
    "MAXsssN",
    "MAXddsN",
    "MAXaasN",
    "MAXssssN",
    "MAXsOH",
    "MAXdO",
    "MAXssO",
    "MAXaaO",
    "MAXsF",
    "MAXsSiH3",
    "MAXssSiH2",
    "MAXsssSiH",
    "MAXssssSi",
    "MAXsPH2",
    "MAXssPH",
    "MAXsssP",
    "MAXdsssP",
    "MAXsssssP",
    "MAXsSH",
    "MAXdS",
    "MAXssS",
    "MAXaaS",
    "MAXdssS",
    "MAXddssS",
    "MAXsCl",
    "MAXsGeH3",
    "MAXssGeH2",
    "MAXsssGeH",
    "MAXssssGe",
    "MAXsAsH2",
    "MAXssAsH",
    "MAXsssAs",
    "MAXsssdAs",
    "MAXsssssAs",
    "MAXsSeH",
    "MAXdSe",
    "MAXssSe",
    "MAXaaSe",
    "MAXdssSe",
    "MAXddssSe",
    "MAXsBr",
    "MAXsSnH3",
    "MAXssSnH2",
    "MAXsssSnH",
    "MAXssssSn",
    "MAXsI",
    "MAXsPbH3",
    "MAXssPbH2",
    "MAXsssPbH",
    "MAXssssPb",
    "MINsLi",
    "MINssBe",
    "MINssssBe",
    "MINssBH",
    "MINsssB",
    "MINssssB",
    "MINsCH3",
    "MINdCH2",
    "MINssCH2",
    "MINtCH",
    "MINdsCH",
    "MINaaCH",
    "MINsssCH",
    "MINddC",
    "MINtsC",
    "MINdssC",
    "MINaasC",
    "MINaaaC",
    "MINssssC",
    "MINsNH3",
    "MINsNH2",
    "MINssNH2",
    "MINdNH",
    "MINssNH",
    "MINaaNH",
    "MINtN",
    "MINsssNH",
    "MINdsN",
    "MINaaN",
    "MINsssN",
    "MINddsN",
    "MINaasN",
    "MINssssN",
    "MINsOH",
    "MINdO",
    "MINssO",
    "MINaaO",
    "MINsF",
    "MINsSiH3",
    "MINssSiH2",
    "MINsssSiH",
    "MINssssSi",
    "MINsPH2",
    "MINssPH",
    "MINsssP",
    "MINdsssP",
    "MINsssssP",
    "MINsSH",
    "MINdS",
    "MINssS",
    "MINaaS",
    "MINdssS",
    "MINddssS",
    "MINsCl",
    "MINsGeH3",
    "MINssGeH2",
    "MINsssGeH",
    "MINssssGe",
    "MINsAsH2",
    "MINssAsH",
    "MINsssAs",
    "MINsssdAs",
    "MINsssssAs",
    "MINsSeH",
    "MINdSe",
    "MINssSe",
    "MINaaSe",
    "MINdssSe",
    "MINddssSe",
    "MINsBr",
    "MINsSnH3",
    "MINssSnH2",
    "MINsssSnH",
    "MINssssSn",
    "MINsI",
    "MINsPbH3",
    "MINssPbH2",
    "MINsssPbH",
    "MINssssPb",
    "ECIndex",
    "ETA_alpha",
    "AETA_alpha",
    "ETA_shape_p",
    "ETA_shape_y",
    "ETA_shape_x",
    "ETA_beta",
    "AETA_beta",
    "ETA_beta_s",
    "AETA_beta_s",
    "ETA_beta_ns",
    "AETA_beta_ns",
    "ETA_beta_ns_d",
    "AETA_beta_ns_d",
    "ETA_eta",
    "AETA_eta",
    "ETA_eta_L",
    "AETA_eta_L",
    "ETA_eta_R",
    "AETA_eta_R",
    "ETA_eta_RL",
    "AETA_eta_RL",
    "ETA_eta_F",
    "AETA_eta_F",
    "ETA_eta_FL",
    "AETA_eta_FL",
    "ETA_eta_B",
    "AETA_eta_B",
    "ETA_eta_BR",
    "AETA_eta_BR",
    "ETA_dAlpha_A",
    "ETA_dAlpha_B",
    "ETA_epsilon_1",
    "ETA_epsilon_2",
    "ETA_epsilon_3",
    "ETA_epsilon_4",
    "ETA_epsilon_5",
    "ETA_dEpsilon_A",
    "ETA_dEpsilon_B",
    "ETA_dEpsilon_C",
    "ETA_dEpsilon_D",
    "ETA_dBeta",
    "AETA_dBeta",
    "ETA_psi_1",
    "ETA_dPsi_A",
    "ETA_dPsi_B",
    "fragCpx",
    "fMF",
    "nHBAcc",
    "nHBDon",
    "IC0",
    "IC1",
    "IC2",
    "IC3",
    "IC4",
    "IC5",
    "TIC0",
    "TIC1",
    "TIC2",
    "TIC3",
    "TIC4",
    "TIC5",
    "SIC0",
    "SIC1",
    "SIC2",
    "SIC3",
    "SIC4",
    "SIC5",
    "BIC0",
    "BIC1",
    "BIC2",
    "BIC3",
    "BIC4",
    "BIC5",
    "CIC0",
    "CIC1",
    "CIC2",
    "CIC3",
    "CIC4",
    "CIC5",
    "MIC0",
    "MIC1",
    "MIC2",
    "MIC3",
    "MIC4",
    "MIC5",
    "ZMIC0",
    "ZMIC1",
    "ZMIC2",
    "ZMIC3",
    "ZMIC4",
    "ZMIC5",
    "Kier1",
    "Kier2",
    "Kier3",
    "Lipinski",
    "GhoseFilter",
    "FilterItLogS",
    "VMcGowan",
    "LabuteASA",
    "PEOE_VSA1",
    "PEOE_VSA2",
    "PEOE_VSA3",
    "PEOE_VSA4",
    "PEOE_VSA5",
    "PEOE_VSA6",
    "PEOE_VSA7",
    "PEOE_VSA8",
    "PEOE_VSA9",
    "PEOE_VSA10",
    "PEOE_VSA11",
    "PEOE_VSA12",
    "PEOE_VSA13",
    "SMR_VSA1",
    "SMR_VSA2",
    "SMR_VSA3",
    "SMR_VSA4",
    "SMR_VSA5",
    "SMR_VSA6",
    "SMR_VSA7",
    "SMR_VSA8",
    "SMR_VSA9",
    "SlogP_VSA1",
    "SlogP_VSA2",
    "SlogP_VSA3",
    "SlogP_VSA4",
    "SlogP_VSA5",
    "SlogP_VSA6",
    "SlogP_VSA7",
    "SlogP_VSA8",
    "SlogP_VSA9",
    "SlogP_VSA10",
    "SlogP_VSA11",
    "EState_VSA1",
    "EState_VSA2",
    "EState_VSA3",
    "EState_VSA4",
    "EState_VSA5",
    "EState_VSA6",
    "EState_VSA7",
    "EState_VSA8",
    "EState_VSA9",
    "EState_VSA10",
    "VSA_EState1",
    "VSA_EState2",
    "VSA_EState3",
    "VSA_EState4",
    "VSA_EState5",
    "VSA_EState6",
    "VSA_EState7",
    "VSA_EState8",
    "VSA_EState9",
    "MDEC-11",
    "MDEC-12",
    "MDEC-13",
    "MDEC-14",
    "MDEC-22",
    "MDEC-23",
    "MDEC-24",
    "MDEC-33",
    "MDEC-34",
    "MDEC-44",
    "MDEO-11",
    "MDEO-12",
    "MDEO-22",
    "MDEN-11",
    "MDEN-12",
    "MDEN-13",
    "MDEN-22",
    "MDEN-23",
    "MDEN-33",
    "MID",
    "AMID",
    "MID_h",
    "AMID_h",
    "MID_C",
    "AMID_C",
    "MID_N",
    "AMID_N",
    "MID_O",
    "AMID_O",
    "MID_X",
    "AMID_X",
    "MPC2",
    "MPC3",
    "MPC4",
    "MPC5",
    "MPC6",
    "MPC7",
    "MPC8",
    "MPC9",
    "MPC10",
    "TMPC10",
    "piPC1",
    "piPC2",
    "piPC3",
    "piPC4",
    "piPC5",
    "piPC6",
    "piPC7",
    "piPC8",
    "piPC9",
    "piPC10",
    "TpiPC10",
    "apol",
    "bpol",
    "nRing",
    "n3Ring",
    "n4Ring",
    "n5Ring",
    "n6Ring",
    "n7Ring",
    "n8Ring",
    "n9Ring",
    "n10Ring",
    "n11Ring",
    "n12Ring",
    "nG12Ring",
    "nHRing",
    "n3HRing",
    "n4HRing",
    "n5HRing",
    "n6HRing",
    "n7HRing",
    "n8HRing",
    "n9HRing",
    "n10HRing",
    "n11HRing",
    "n12HRing",
    "nG12HRing",
    "naRing",
    "n3aRing",
    "n4aRing",
    "n5aRing",
    "n6aRing",
    "n7aRing",
    "n8aRing",
    "n9aRing",
    "n10aRing",
    "n11aRing",
    "n12aRing",
    "nG12aRing",
    "naHRing",
    "n3aHRing",
    "n4aHRing",
    "n5aHRing",
    "n6aHRing",
    "n7aHRing",
    "n8aHRing",
    "n9aHRing",
    "n10aHRing",
    "n11aHRing",
    "n12aHRing",
    "nG12aHRing",
    "nARing",
    "n3ARing",
    "n4ARing",
    "n5ARing",
    "n6ARing",
    "n7ARing",
    "n8ARing",
    "n9ARing",
    "n10ARing",
    "n11ARing",
    "n12ARing",
    "nG12ARing",
    "nAHRing",
    "n3AHRing",
    "n4AHRing",
    "n5AHRing",
    "n6AHRing",
    "n7AHRing",
    "n8AHRing",
    "n9AHRing",
    "n10AHRing",
    "n11AHRing",
    "n12AHRing",
    "nG12AHRing",
    "nFRing",
    "n4FRing",
    "n5FRing",
    "n6FRing",
    "n7FRing",
    "n8FRing",
    "n9FRing",
    "n10FRing",
    "n11FRing",
    "n12FRing",
    "nG12FRing",
    "nFHRing",
    "n4FHRing",
    "n5FHRing",
    "n6FHRing",
    "n7FHRing",
    "n8FHRing",
    "n9FHRing",
    "n10FHRing",
    "n11FHRing",
    "n12FHRing",
    "nG12FHRing",
    "nFaRing",
    "n4FaRing",
    "n5FaRing",
    "n6FaRing",
    "n7FaRing",
    "n8FaRing",
    "n9FaRing",
    "n10FaRing",
    "n11FaRing",
    "n12FaRing",
    "nG12FaRing",
    "nFaHRing",
    "n4FaHRing",
    "n5FaHRing",
    "n6FaHRing",
    "n7FaHRing",
    "n8FaHRing",
    "n9FaHRing",
    "n10FaHRing",
    "n11FaHRing",
    "n12FaHRing",
    "nG12FaHRing",
    "nFARing",
    "n4FARing",
    "n5FARing",
    "n6FARing",
    "n7FARing",
    "n8FARing",
    "n9FARing",
    "n10FARing",
    "n11FARing",
    "n12FARing",
    "nG12FARing",
    "nFAHRing",
    "n4FAHRing",
    "n5FAHRing",
    "n6FAHRing",
    "n7FAHRing",
    "n8FAHRing",
    "n9FAHRing",
    "n10FAHRing",
    "n11FAHRing",
    "n12FAHRing",
    "nG12FAHRing",
    "nRot",
    "RotRatio",
    "SLogP",
    "SMR",
    "TopoPSA(NO)",
    "TopoPSA",
    "GGI1",
    "GGI2",
    "GGI3",
    "GGI4",
    "GGI5",
    "GGI6",
    "GGI7",
    "GGI8",
    "GGI9",
    "GGI10",
    "JGI1",
    "JGI2",
    "JGI3",
    "JGI4",
    "JGI5",
    "JGI6",
    "JGI7",
    "JGI8",
    "JGI9",
    "JGI10",
    "JGT10",
    "Diameter",
    "Radius",
    "TopoShapeIndex",
    "PetitjeanIndex",
    "Vabc",
    "VAdjMat",
    "MWC01",
    "MWC02",
    "MWC03",
    "MWC04",
    "MWC05",
    "MWC06",
    "MWC07",
    "MWC08",
    "MWC09",
    "MWC10",
    "TMWC10",
    "SRW02",
    "SRW03",
    "SRW04",
    "SRW05",
    "SRW06",
    "SRW07",
    "SRW08",
    "SRW09",
    "SRW10",
    "TSRW10",
    "MW",
    "AMW",
    "WPath",
    "WPol",
    "Zagreb1",
    "Zagreb2",
    "mZagreb1",
    "mZagreb2",
]


SUBSET_947 = [
    "ABC",
    "ABCGG",
    "nBase",
    "SpAbs_A",
    "SpMax_A",
    "SpDiam_A",
    "SpAD_A",
    "SpMAD_A",
    "LogEE_A",
    "VE1_A",
    "VE2_A",
    "VE3_A",
    "VR1_A",
    "VR2_A",
    "VR3_A",
    "nAromAtom",
    "nAromBond",
    "nAtom",
    "nHeavyAtom",
    "nHetero",
    "nH",
    "nC",
    "nN",
    "nO",
    "ATS0dv",
    "ATS1dv",
    "ATS2dv",
    "ATS3dv",
    "ATS0d",
    "ATS1d",
    "ATS2d",
    "ATS3d",
    "ATS4d",
    "ATS5d",
    "ATS0s",
    "ATS1s",
    "ATS2s",
    "ATS3s",
    "ATS4s",
    "ATS5s",
    "ATS0Z",
    "ATS1Z",
    "ATS2Z",
    "ATS3Z",
    "ATS4Z",
    "ATS5Z",
    "ATS0m",
    "ATS1m",
    "ATS2m",
    "ATS3m",
    "ATS4m",
    "ATS5m",
    "ATS0v",
    "ATS1v",
    "ATS2v",
    "ATS3v",
    "ATS4v",
    "ATS5v",
    "ATS0se",
    "ATS1se",
    "ATS2se",
    "ATS3se",
    "ATS4se",
    "ATS5se",
    "ATS0pe",
    "ATS1pe",
    "ATS2pe",
    "ATS3pe",
    "ATS4pe",
    "ATS5pe",
    "ATS0are",
    "ATS1are",
    "ATS2are",
    "ATS3are",
    "ATS4are",
    "ATS5are",
    "ATS0p",
    "ATS1p",
    "ATS2p",
    "ATS3p",
    "ATS4p",
    "ATS5p",
    "ATS0i",
    "ATS1i",
    "ATS2i",
    "ATS3i",
    "ATS4i",
    "ATS5i",
    "AATS0dv",
    "AATS1dv",
    "AATS2dv",
    "AATS3dv",
    "AATS0d",
    "AATS1d",
    "AATS2d",
    "AATS3d",
    "AATS4d",
    "AATS0s",
    "AATS1s",
    "AATS2s",
    "AATS3s",
    "AATS4s",
    "AATS0Z",
    "AATS1Z",
    "AATS2Z",
    "AATS3Z",
    "AATS4Z",
    "AATS0m",
    "AATS1m",
    "AATS2m",
    "AATS3m",
    "AATS4m",
    "AATS0v",
    "AATS1v",
    "AATS2v",
    "AATS3v",
    "AATS4v",
    "AATS5v",
    "AATS0se",
    "AATS1se",
    "AATS2se",
    "AATS3se",
    "AATS4se",
    "AATS5se",
    "AATS0pe",
    "AATS1pe",
    "AATS2pe",
    "AATS3pe",
    "AATS4pe",
    "AATS5pe",
    "AATS0are",
    "AATS1are",
    "AATS2are",
    "AATS3are",
    "AATS4are",
    "AATS5are",
    "AATS0p",
    "AATS1p",
    "AATS2p",
    "AATS3p",
    "AATS4p",
    "AATS5p",
    "AATS0i",
    "AATS1i",
    "AATS2i",
    "AATS3i",
    "AATS4i",
    "AATS5i",
    "ATSC0c",
    "ATSC1c",
    "ATSC2c",
    "ATSC3c",
    "ATSC4c",
    "ATSC5c",
    "ATSC0dv",
    "ATSC1dv",
    "ATSC2dv",
    "ATSC3dv",
    "ATSC4dv",
    "ATSC5dv",
    "ATSC0d",
    "ATSC1d",
    "ATSC2d",
    "ATSC3d",
    "ATSC4d",
    "ATSC5d",
    "ATSC0s",
    "ATSC1s",
    "ATSC2s",
    "ATSC3s",
    "ATSC4s",
    "ATSC5s",
    "ATSC0Z",
    "ATSC1Z",
    "ATSC2Z",
    "ATSC3Z",
    "ATSC4Z",
    "ATSC5Z",
    "ATSC0m",
    "ATSC1m",
    "ATSC2m",
    "ATSC3m",
    "ATSC4m",
    "ATSC5m",
    "ATSC0v",
    "ATSC1v",
    "ATSC2v",
    "ATSC3v",
    "ATSC4v",
    "ATSC5v",
    "ATSC0se",
    "ATSC1se",
    "ATSC2se",
    "ATSC3se",
    "ATSC4se",
    "ATSC5se",
    "ATSC0pe",
    "ATSC1pe",
    "ATSC2pe",
    "ATSC3pe",
    "ATSC4pe",
    "ATSC5pe",
    "ATSC0are",
    "ATSC1are",
    "ATSC2are",
    "ATSC3are",
    "ATSC4are",
    "ATSC5are",
    "ATSC0p",
    "ATSC1p",
    "ATSC2p",
    "ATSC3p",
    "ATSC4p",
    "ATSC5p",
    "ATSC0i",
    "ATSC1i",
    "ATSC2i",
    "ATSC3i",
    "ATSC4i",
    "ATSC5i",
    "AATSC0c",
    "AATSC1c",
    "AATSC2c",
    "AATSC3c",
    "AATSC4c",
    "AATSC5c",
    "AATSC0dv",
    "AATSC1dv",
    "AATSC2dv",
    "AATSC3dv",
    "AATSC4dv",
    "AATSC5dv",
    "AATSC0d",
    "AATSC1d",
    "AATSC2d",
    "AATSC3d",
    "AATSC4d",
    "AATSC5d",
    "AATSC0s",
    "AATSC1s",
    "AATSC2s",
    "AATSC3s",
    "AATSC4s",
    "AATSC5s",
    "AATSC0Z",
    "AATSC1Z",
    "AATSC2Z",
    "AATSC3Z",
    "AATSC4Z",
    "AATSC5Z",
    "AATSC0m",
    "AATSC1m",
    "AATSC2m",
    "AATSC3m",
    "AATSC4m",
    "AATSC5m",
    "AATSC0v",
    "AATSC1v",
    "AATSC2v",
    "AATSC3v",
    "AATSC4v",
    "AATSC5v",
    "AATSC0se",
    "AATSC1se",
    "AATSC2se",
    "AATSC3se",
    "AATSC4se",
    "AATSC5se",
    "AATSC0pe",
    "AATSC1pe",
    "AATSC2pe",
    "AATSC3pe",
    "AATSC4pe",
    "AATSC5pe",
    "AATSC0are",
    "AATSC1are",
    "AATSC2are",
    "AATSC3are",
    "AATSC4are",
    "AATSC5are",
    "AATSC0p",
    "AATSC1p",
    "AATSC2p",
    "AATSC3p",
    "AATSC4p",
    "AATSC5p",
    "AATSC0i",
    "AATSC1i",
    "AATSC2i",
    "AATSC3i",
    "AATSC4i",
    "AATSC5i",
    "MATS1c",
    "MATS2c",
    "MATS3c",
    "MATS4c",
    "MATS5c",
    "MATS1dv",
    "MATS2dv",
    "MATS3dv",
    "MATS4dv",
    "MATS5dv",
    "MATS1d",
    "MATS2d",
    "MATS3d",
    "MATS4d",
    "MATS5d",
    "MATS1s",
    "MATS2s",
    "MATS3s",
    "MATS4s",
    "MATS5s",
    "MATS1Z",
    "MATS2Z",
    "MATS3Z",
    "MATS4Z",
    "MATS5Z",
    "MATS1m",
    "MATS2m",
    "MATS3m",
    "MATS4m",
    "MATS5m",
    "MATS1v",
    "MATS2v",
    "MATS3v",
    "MATS4v",
    "MATS5v",
    "MATS1se",
    "MATS2se",
    "MATS3se",
    "MATS4se",
    "MATS5se",
    "MATS1pe",
    "MATS2pe",
    "MATS3pe",
    "MATS4pe",
    "MATS5pe",
    "MATS1are",
    "MATS2are",
    "MATS3are",
    "MATS4are",
    "MATS5are",
    "MATS1p",
    "MATS2p",
    "MATS3p",
    "MATS4p",
    "MATS5p",
    "MATS1i",
    "MATS2i",
    "MATS3i",
    "MATS4i",
    "MATS5i",
    "GATS1c",
    "GATS2c",
    "GATS3c",
    "GATS4c",
    "GATS5c",
    "GATS1dv",
    "GATS2dv",
    "GATS3dv",
    "GATS4dv",
    "GATS1d",
    "GATS2d",
    "GATS3d",
    "GATS4d",
    "GATS1s",
    "GATS2s",
    "GATS3s",
    "GATS4s",
    "GATS1Z",
    "GATS2Z",
    "GATS3Z",
    "GATS4Z",
    "GATS1m",
    "GATS2m",
    "GATS3m",
    "GATS4m",
    "GATS1v",
    "GATS2v",
    "GATS3v",
    "GATS4v",
    "GATS1se",
    "GATS2se",
    "GATS3se",
    "GATS4se",
    "GATS1pe",
    "GATS2pe",
    "GATS3pe",
    "GATS4pe",
    "GATS1are",
    "GATS2are",
    "GATS3are",
    "GATS4are",
    "GATS1p",
    "GATS2p",
    "GATS3p",
    "GATS4p",
    "GATS1i",
    "GATS2i",
    "GATS3i",
    "GATS4i",
    "BCUTc-1h",
    "BCUTc-1l",
    "BCUTdv-1h",
    "BCUTdv-1l",
    "BCUTd-1h",
    "BCUTd-1l",
    "BCUTs-1h",
    "BCUTs-1l",
    "BCUTZ-1h",
    "BCUTZ-1l",
    "BCUTm-1h",
    "BCUTm-1l",
    "BCUTv-1h",
    "BCUTv-1l",
    "BCUTse-1h",
    "BCUTse-1l",
    "BCUTpe-1h",
    "BCUTpe-1l",
    "BCUTare-1h",
    "BCUTare-1l",
    "BCUTp-1h",
    "BCUTp-1l",
    "BCUTi-1h",
    "BCUTi-1l",
    "BalabanJ",
    "SpAbs_DzZ",
    "SpMax_DzZ",
    "SpDiam_DzZ",
    "SpAD_DzZ",
    "SpMAD_DzZ",
    "LogEE_DzZ",
    "SM1_DzZ",
    "VE1_DzZ",
    "VE2_DzZ",
    "VE3_DzZ",
    "VR1_DzZ",
    "VR2_DzZ",
    "VR3_DzZ",
    "SpAbs_Dzm",
    "SpMax_Dzm",
    "SpDiam_Dzm",
    "SpAD_Dzm",
    "SpMAD_Dzm",
    "LogEE_Dzm",
    "SM1_Dzm",
    "VE1_Dzm",
    "VE2_Dzm",
    "VE3_Dzm",
    "VR1_Dzm",
    "VR2_Dzm",
    "VR3_Dzm",
    "SpAbs_Dzv",
    "SpMax_Dzv",
    "SpDiam_Dzv",
    "SpAD_Dzv",
    "SpMAD_Dzv",
    "LogEE_Dzv",
    "SM1_Dzv",
    "VE1_Dzv",
    "VE2_Dzv",
    "VE3_Dzv",
    "VR1_Dzv",
    "VR2_Dzv",
    "VR3_Dzv",
    "SpAbs_Dzse",
    "SpMax_Dzse",
    "SpDiam_Dzse",
    "SpAD_Dzse",
    "SpMAD_Dzse",
    "LogEE_Dzse",
    "SM1_Dzse",
    "VE1_Dzse",
    "VE2_Dzse",
    "VE3_Dzse",
    "VR1_Dzse",
    "VR2_Dzse",
    "VR3_Dzse",
    "SpAbs_Dzpe",
    "SpMax_Dzpe",
    "SpDiam_Dzpe",
    "SpAD_Dzpe",
    "SpMAD_Dzpe",
    "LogEE_Dzpe",
    "SM1_Dzpe",
    "VE1_Dzpe",
    "VE2_Dzpe",
    "VE3_Dzpe",
    "VR1_Dzpe",
    "VR2_Dzpe",
    "VR3_Dzpe",
    "SpAbs_Dzare",
    "SpMax_Dzare",
    "SpDiam_Dzare",
    "SpAD_Dzare",
    "SpMAD_Dzare",
    "LogEE_Dzare",
    "SM1_Dzare",
    "VE1_Dzare",
    "VE2_Dzare",
    "VE3_Dzare",
    "VR1_Dzare",
    "VR2_Dzare",
    "VR3_Dzare",
    "SpAbs_Dzp",
    "SpMax_Dzp",
    "SpDiam_Dzp",
    "SpAD_Dzp",
    "SpMAD_Dzp",
    "LogEE_Dzp",
    "SM1_Dzp",
    "VE1_Dzp",
    "VE2_Dzp",
    "VE3_Dzp",
    "VR1_Dzp",
    "VR2_Dzp",
    "VR3_Dzp",
    "SpAbs_Dzi",
    "SpMax_Dzi",
    "SpDiam_Dzi",
    "SpAD_Dzi",
    "SpMAD_Dzi",
    "LogEE_Dzi",
    "SM1_Dzi",
    "VE1_Dzi",
    "VE2_Dzi",
    "VE3_Dzi",
    "VR1_Dzi",
    "VR2_Dzi",
    "VR3_Dzi",
    "BertzCT",
    "nBonds",
    "nBondsO",
    "nBondsS",
    "nBondsD",
    "nBondsT",
    "nBondsA",
    "nBondsM",
    "nBondsKS",
    "nBondsKD",
    "RNCG",
    "RPCG",
    "C1SP1",
    "C2SP1",
    "C1SP2",
    "C2SP2",
    "C1SP3",
    "C2SP3",
    "C3SP3",
    "C4SP3",
    "HybRatio",
    "FCSP3",
    "Xch-3d",
    "Xch-4d",
    "Xch-5d",
    "Xch-3dv",
    "Xch-4dv",
    "Xch-5dv",
    "Xc-3d",
    "Xc-4d",
    "Xc-3dv",
    "Xc-4dv",
    "Xpc-4d",
    "Xpc-4dv",
    "Xp-0d",
    "Xp-1d",
    "Xp-2d",
    "Xp-3d",
    "Xp-4d",
    "AXp-0d",
    "AXp-1d",
    "AXp-2d",
    "AXp-3d",
    "AXp-4d",
    "Xp-0dv",
    "Xp-1dv",
    "Xp-2dv",
    "Xp-3dv",
    "Xp-4dv",
    "AXp-0dv",
    "AXp-1dv",
    "AXp-2dv",
    "AXp-3dv",
    "AXp-4dv",
    "SZ",
    "Sm",
    "Sv",
    "Sse",
    "Spe",
    "Sare",
    "Sp",
    "Si",
    "MZ",
    "Mm",
    "Mv",
    "Mse",
    "Mpe",
    "Mare",
    "Mp",
    "Mi",
    "SpAbs_Dt",
    "SpMax_Dt",
    "SpDiam_Dt",
    "SpAD_Dt",
    "SpMAD_Dt",
    "LogEE_Dt",
    "VE1_Dt",
    "VE2_Dt",
    "VE3_Dt",
    "VR1_Dt",
    "VR2_Dt",
    "VR3_Dt",
    "DetourIndex",
    "SpAbs_D",
    "SpMax_D",
    "SpDiam_D",
    "SpAD_D",
    "SpMAD_D",
    "LogEE_D",
    "VE1_D",
    "VE2_D",
    "VE3_D",
    "VR1_D",
    "VR2_D",
    "VR3_D",
    "NsCH3",
    "NdCH2",
    "NssCH2",
    "NtCH",
    "NdsCH",
    "NaaCH",
    "NsssCH",
    "NtsC",
    "NdssC",
    "NssssC",
    "NsNH2",
    "NdNH",
    "NssNH",
    "NaaNH",
    "NtN",
    "NdsN",
    "NaaN",
    "NsssN",
    "NsOH",
    "NdO",
    "NssO",
    "NaaO",
    "SsCH3",
    "SdCH2",
    "SssCH2",
    "StCH",
    "SdsCH",
    "SaaCH",
    "SsssCH",
    "StsC",
    "SdssC",
    "SssssC",
    "SsNH2",
    "SdNH",
    "SssNH",
    "SaaNH",
    "StN",
    "SdsN",
    "SaaN",
    "SsssN",
    "SsOH",
    "SdO",
    "SssO",
    "SaaO",
    "MAXsCH3",
    "MAXssCH2",
    "MAXtCH",
    "MAXdsCH",
    "MAXaaCH",
    "MAXsssCH",
    "MAXtsC",
    "MAXdssC",
    "MAXssssC",
    "MAXsNH2",
    "MAXdNH",
    "MAXssNH",
    "MAXaaNH",
    "MAXtN",
    "MAXaaN",
    "MAXsssN",
    "MAXsOH",
    "MAXdO",
    "MAXssO",
    "MAXaaO",
    "MINsCH3",
    "MINssCH2",
    "MINtCH",
    "MINdsCH",
    "MINaaCH",
    "MINsssCH",
    "MINtsC",
    "MINdssC",
    "MINssssC",
    "MINsNH2",
    "MINdNH",
    "MINssNH",
    "MINaaNH",
    "MINtN",
    "MINaaN",
    "MINsssN",
    "MINsOH",
    "MINdO",
    "MINssO",
    "MINaaO",
    "ECIndex",
    "ETA_alpha",
    "AETA_alpha",
    "ETA_shape_p",
    "ETA_shape_y",
    "ETA_shape_x",
    "ETA_beta",
    "AETA_beta",
    "ETA_beta_s",
    "AETA_beta_s",
    "ETA_beta_ns",
    "AETA_beta_ns",
    "ETA_eta",
    "AETA_eta",
    "ETA_eta_L",
    "AETA_eta_L",
    "ETA_eta_R",
    "AETA_eta_R",
    "ETA_eta_RL",
    "AETA_eta_RL",
    "ETA_eta_F",
    "AETA_eta_F",
    "ETA_eta_FL",
    "AETA_eta_FL",
    "ETA_eta_B",
    "AETA_eta_B",
    "ETA_eta_BR",
    "AETA_eta_BR",
    "ETA_dAlpha_B",
    "ETA_epsilon_1",
    "ETA_epsilon_2",
    "ETA_epsilon_3",
    "ETA_epsilon_4",
    "ETA_epsilon_5",
    "ETA_dEpsilon_A",
    "ETA_dEpsilon_B",
    "ETA_dEpsilon_C",
    "ETA_dEpsilon_D",
    "ETA_dBeta",
    "AETA_dBeta",
    "ETA_psi_1",
    "ETA_dPsi_A",
    "ETA_dPsi_B",
    "fragCpx",
    "fMF",
    "nHBAcc",
    "nHBDon",
    "IC0",
    "IC1",
    "IC2",
    "IC3",
    "IC4",
    "IC5",
    "TIC0",
    "TIC1",
    "TIC2",
    "TIC3",
    "TIC4",
    "TIC5",
    "SIC0",
    "SIC1",
    "SIC2",
    "SIC3",
    "SIC4",
    "SIC5",
    "BIC0",
    "BIC1",
    "BIC2",
    "BIC3",
    "BIC4",
    "BIC5",
    "CIC0",
    "CIC1",
    "CIC2",
    "CIC3",
    "CIC4",
    "CIC5",
    "MIC0",
    "MIC1",
    "MIC2",
    "MIC3",
    "MIC4",
    "MIC5",
    "ZMIC0",
    "ZMIC1",
    "ZMIC2",
    "ZMIC3",
    "ZMIC4",
    "ZMIC5",
    "Kier1",
    "Kier2",
    "Kier3",
    "FilterItLogS",
    "VMcGowan",
    "LabuteASA",
    "PEOE_VSA1",
    "PEOE_VSA2",
    "PEOE_VSA3",
    "PEOE_VSA4",
    "PEOE_VSA5",
    "PEOE_VSA6",
    "PEOE_VSA7",
    "PEOE_VSA8",
    "PEOE_VSA9",
    "PEOE_VSA10",
    "PEOE_VSA11",
    "PEOE_VSA12",
    "PEOE_VSA13",
    "SMR_VSA1",
    "SMR_VSA2",
    "SMR_VSA3",
    "SMR_VSA4",
    "SMR_VSA5",
    "SMR_VSA6",
    "SMR_VSA7",
    "SMR_VSA9",
    "SlogP_VSA1",
    "SlogP_VSA2",
    "SlogP_VSA3",
    "SlogP_VSA4",
    "SlogP_VSA5",
    "SlogP_VSA6",
    "SlogP_VSA10",
    "EState_VSA1",
    "EState_VSA2",
    "EState_VSA3",
    "EState_VSA4",
    "EState_VSA5",
    "EState_VSA6",
    "EState_VSA7",
    "EState_VSA8",
    "EState_VSA9",
    "EState_VSA10",
    "VSA_EState1",
    "VSA_EState2",
    "VSA_EState3",
    "VSA_EState4",
    "VSA_EState5",
    "VSA_EState6",
    "VSA_EState7",
    "VSA_EState8",
    "VSA_EState9",
    "MDEC-11",
    "MDEC-12",
    "MDEC-13",
    "MDEC-14",
    "MDEC-22",
    "MDEC-23",
    "MDEC-24",
    "MDEO-12",
    "MDEN-11",
    "MID",
    "AMID",
    "MID_h",
    "AMID_h",
    "MID_C",
    "AMID_C",
    "MID_N",
    "AMID_N",
    "MID_O",
    "AMID_O",
    "MPC2",
    "MPC3",
    "MPC4",
    "TMPC10",
    "piPC1",
    "piPC2",
    "piPC3",
    "piPC4",
    "TpiPC10",
    "apol",
    "bpol",
    "nRing",
    "n3Ring",
    "n4Ring",
    "n5Ring",
    "nHRing",
    "n3HRing",
    "n4HRing",
    "n5HRing",
    "naRing",
    "n5aRing",
    "naHRing",
    "n5aHRing",
    "nARing",
    "n3ARing",
    "n4ARing",
    "nAHRing",
    "n3AHRing",
    "n4AHRing",
    "nRot",
    "RotRatio",
    "SLogP",
    "SMR",
    "TopoPSA(NO)",
    "TopoPSA",
    "GGI1",
    "GGI2",
    "GGI3",
    "JGI1",
    "JGI2",
    "JGI3",
    "JGT10",
    "Diameter",
    "Radius",
    "TopoShapeIndex",
    "PetitjeanIndex",
    "Vabc",
    "VAdjMat",
    "MWC01",
    "MWC02",
    "MWC03",
    "MWC04",
    "MWC05",
    "MWC06",
    "MWC07",
    "MWC08",
    "MWC09",
    "MWC10",
    "TMWC10",
    "SRW02",
    "SRW03",
    "SRW04",
    "SRW05",
    "SRW06",
    "SRW07",
    "SRW08",
    "SRW09",
    "SRW10",
    "TSRW10",
    "MW",
    "AMW",
    "WPath",
    "WPol",
    "Zagreb1",
    "Zagreb2",
    "mZagreb1",
    "mZagreb2",
]

descriptors_lookup = dict(
    optimized=SUBSET_947,
    all=ALL_2D,
)



# given a filepath and some target columns,
# retrieve the data as numpy arrays


import pandas as pd
from sklearn.feature_selection import VarianceThreshold

from fastprop.defaults import init_logger

logger = init_logger(__name__)


def preprocess(descriptors, zero_variance_drop=False, colinear_drop=False):
    # drop missing features - if this is direct output from mordred, missing values are
    # strings describing why they are missing. If this is output from fastprop.utils.load_daved_desc
    # the missing descriptors are nan. To deal with the former case, force all str->nan
    descriptors: pd.DataFrame
    descriptors = descriptors.apply(pd.to_numeric, errors="coerce")
    descriptors = descriptors.dropna(axis=1, how="all")

    if zero_variance_drop:
        # drop invariant features
        var_scaler = VarianceThreshold(threshold=0).set_output(transform="pandas")
        # exclude from the returned feature scalers - the columns which should be dropped are done
        # automatically by not calculating them in the first place during prediction.
        descriptors = var_scaler.fit_transform(descriptors)
        logger.info(f"size after invariant feature removal: {descriptors.shape}")

    if colinear_drop:
        raise NotImplementedError("TODO")

    logger.info(f"Size after preprocessing: {descriptors.shape}")

    return descriptors



import datetime
import os
from multiprocessing import Pool
from pathlib import Path
from time import perf_counter
from typing import Literal

import numpy as np
import pandas as pd
import psutil
from mordred import Calculator

from fastprop.defaults import init_logger

from .descriptor_lists import descriptors_lookup
from .load_data import load_saved_desc
from .select_descriptors import mordred_descriptors_from_strings

logger = init_logger(__name__)


# mordred tried to avoid instantiating multiple Calculator classes, which makes
# the parallelism slower but more memory efficient. We do this manually:
def _f(in_tuple):
    mordred_calc = Calculator(in_tuple[2], ignore_3D=in_tuple[3])
    mordred_descs = np.array(list(mordred_calc.map(in_tuple[0], nproc=1, quiet=in_tuple[1])))
    return mordred_descs


def calculate_mordred_desciptors(descriptors, rdkit_mols, n_procs, strategy: Literal["fast", "low-memory"] = "fast", ignore_3d=True):
    """Wraps the mordred descriptor calculator.

    Args:
        descriptors (Mordred descriptor instances): Descriptors to calculate
        rdkit_mols (list[rdkit mols]): List of RDKit molecules.
        n_procs (int): Number of parallel processes.
        strategy (Literal["fast", "low-memory", optional): Parallelization strategy. Defaults to "fast".
        ignore_3d (bool, optional): Include 3D descriptors, if in given list. Defaults to True.

    Raises:
        RuntimeError: Invalid choice of parallel strategy.

    Returns:
        np.array: Calculated descriptors.
    """
    start = perf_counter()
    # descriptors should be a list of mordred descriptors classes
    if strategy not in {"fast", "low-memory"}:
        raise RuntimeError(f"Strategy {strategy} not supported, only 'fast' and 'low-memory'.")

    mordred_descs = None
    logger.info(f"Calculating descriptors using {strategy=}")
    if strategy == "fast":
        # higher level parallelism - uses more memory
        # TODO: subdivide batches further to avoid large communication bottleneck after all descriptors are calculated
        batches = np.array_split(rdkit_mols, n_procs)
        # let the root process show a progress bar, since array split will make
        # that one the largest
        # convert to starmap to avoid having to duplicate the list of descriptor classes - too much communication (?)
        to_procs = [(batch, bool(i), descriptors, ignore_3d) for i, batch in enumerate(batches)]
        with Pool(n_procs) as p:
            mordred_descs = np.vstack(p.map(_f, to_procs, 1))
    else:
        # mordred parallelism
        mordred_calc = Calculator(descriptors, ignore_3D=ignore_3d)
        mordred_descs = np.array(list(mordred_calc.map(rdkit_mols, nproc=psutil.cpu_count(logical=True), quiet=False)))
    logger.info(f"Descriptor calculation complete, elapsed time: {str(datetime.timedelta(seconds=perf_counter() - start))}")
    return mordred_descs


def _get_descs(precomputed, input_file, output_directory, descriptors, enable_cache, mols, as_df=False):
    """Loads descriptors according to the user-specified configuration.

    This is a 'hidden' function since the caching logic is specific to fastprop.

    Args:
        precomputed (str): Use precomputed descriptors if str is.
        input_file (str): Filepath of input data.
        output_directory (str): Destination directory for caching.
        descriptors (list): fastprop set of descriptors to calculate.
        enable_cache (bool): Allow/disallow caching mechanism.
        mols (list): RDKit molecules.
    """
    descs = None
    if precomputed:
        del mols
        logger.info(f"Loading precomputed descriptors from {precomputed}.")
        descs = load_saved_desc(precomputed)
    else:
        in_name = Path(input_file).stem
        # cached descriptors, which contains (1) cached (2) source filename (3) types of descriptors (4) timestamp when file was last touched
        cache_file = os.path.join(output_directory, "cached_" + in_name + "_" + descriptors + "_" + str(int(os.stat(input_file).st_ctime)) + ".csv")

        if os.path.exists(cache_file) and enable_cache:
            logger.info(f"Found cached descriptor data at {cache_file}, loading instead of recalculating.")
            descs = load_saved_desc(cache_file)
        else:
            d2c = mordred_descriptors_from_strings(descriptors_lookup[descriptors])
            # use all the cpus available
            descs = calculate_mordred_desciptors(d2c, mols, psutil.cpu_count(logical=False), "fast")
            # cache these
            if enable_cache:
                d = pd.DataFrame(descs)
                d.to_csv(cache_file)
                logger.info(f"Cached descriptors to {cache_file}.")
    if as_df:
        if precomputed:  # specifying descriptors' names not required
            return pd.DataFrame(data=descs)
        return pd.DataFrame(data=descs, columns=descriptors_lookup[descriptors])
    return descs



import os
import pickle

import numpy as np
import pandas as pd
import yaml
from rdkit import Chem

from fastprop.defaults import init_logger
from fastprop.utils import calculate_mordred_desciptors
from fastprop.utils.select_descriptors import mordred_descriptors_from_strings

from .fastprop_core import fastprop

logger = init_logger(__name__)


def _safe_yaml_open(dir, fname):
    try:
        with open(os.path.join(dir, fname)) as file:
            return yaml.safe_load(file)
    except FileNotFoundError:
        logger.error(f"checkpoints directory is missing '{fname}'. Re-execute training.")


def predict_fastprop(checkpoints_dir, smiles, input_file, output=None):
    """Prediction CLI.

    Loads a model and runs inference on the input.

    Args:
        checkpoints_dir (str): 'checkpoints' directory from a previous fastprop train.
        smiles (str or list[str]): SMILES strings for prediction.
        input_file (str): Input file containing only SMILES strings for prediction.
        output (str or None): Either save to a file or just print result.
    """
    if input_file:
        raise NotImplementedError("TODO: please pass as command line options, loading from file is a WIP")
    if type(smiles) is str:
        smiles = [smiles]
    checkpoint_dir_contents = os.listdir(checkpoints_dir)
    config_dict = _safe_yaml_open(checkpoints_dir, "fastprop_config.yml")
    descs = calculate_mordred_desciptors(
        mordred_descriptors_from_strings(config_dict["descriptors"]),
        [Chem.MolFromSmiles(i) for i in smiles],
        n_procs=0,  # ignored for "strategy='low-memory'"
        strategy="low-memory",
    )
    all_models = []
    for checkpoint in checkpoint_dir_contents:
        if not checkpoint.endswith(".ckpt"):
            continue
        model = fastprop.load_from_checkpoint(
            os.path.join(checkpoints_dir, checkpoint),
            cleaned_data=None,
            targets=None,
            smiles=None,
        )
        scalers = _safe_yaml_open(checkpoints_dir, f"repetition_{checkpoint.split('-')[1]}_scalers.yml")
        model.target_scaler = pickle.loads(scalers["target_scaler"])
        model.mean_imputer = pickle.loads(scalers["mean_imputer"])
        model.feature_scaler = pickle.loads(scalers["feature_scaler"])
        all_models.append(model)
    # axis: contents
    # 0: smiles
    # 1: predictions
    # 2: per-model
    all_predictions = np.stack([model.predict_step(descs) for model in all_models], axis=2)
    perf = np.mean(all_predictions, axis=2)
    err = np.std(all_predictions, axis=2)
    # interleave the columns of these arrays, thanks stackoverflow.com/a/75519265
    res = np.empty((len(perf), perf.shape[1] * 2), dtype=perf.dtype)
    res[:, 0::2] = perf
    res[:, 1::2] = err
    column_names = []
    for target in config_dict["targets"]:
        column_names.extend([target, target + "_stdev"])
    out = pd.DataFrame(res, columns=column_names, index=smiles)
    if output is None:
        print("\n", out)
    else:
        out.to_csv(output)



import argparse
import datetime
import sys
from importlib.metadata import version
from time import perf_counter

import yaml

from fastprop import (
    DEFAULT_TRAINING_CONFIG,
    hopt_fastprop,
    predict_fastprop,
    shap_fastprop,
    train_fastprop,
)
from fastprop.defaults import init_logger
from fastprop.utils import validate_config

logger = init_logger(__name__)


def main():
    cli_start = perf_counter()
    parser = argparse.ArgumentParser(description="fastprop command line interface - try 'fastprop subcommand --help'")

    parser.add_argument("-v", "--version", help="print version and exit", action="store_true")

    subparsers = parser.add_subparsers(dest="subcommand")

    train_subparser = subparsers.add_parser("train")
    train_subparser.add_argument("config_file", nargs="?", help="YAML configuration file")
    train_subparser.add_argument("-od", "--output-directory", help="directory for fastprop output")
    # featurization
    train_subparser.add_argument("-if", "--input-file", help="csv of SMILES and targets")
    train_subparser.add_argument("-tc", "--target-columns", nargs="+", help="column name(s) for target(s)")
    train_subparser.add_argument("-sc", "--smiles-column", help="column name for SMILES")
    train_subparser.add_argument("-d", "--descriptors", help="descriptors to calculate (one of all, optimized, smallest, or search)")
    train_subparser.add_argument("-ec", "--enable-cache", type=bool, help="allow saving and loading of cached descriptors")
    train_subparser.add_argument("-p", "--precomputed", help="precomputed descriptors from fastprop or mordred")

    # preprocessing
    train_subparser.add_argument("-r", "--rescaling", type=bool, help="rescale descriptors between 0 and 1 (default to True)")
    train_subparser.add_argument("-zvd", "--zero-variance-drop", type=bool, help="drop zero variance descriptors (defaults to True)")
    train_subparser.add_argument("-cd", "--colinear-drop", type=bool, help="drop colinear descriptors (defaults to False)")

    # training
    train_subparser.add_argument("-op", "--optimize", action="store_true", help="run hyperparameter optimization", default=False)
    train_subparser.add_argument("-fl", "--fnn-layers", type=int, help="number of fnn layers")
    train_subparser.add_argument("-lr", "--learning-rate", type=float, help="learning rate")
    train_subparser.add_argument("-bs", "--batch-size", type=int, help="batch size")
    train_subparser.add_argument("-ne", "--number-epochs", type=int, help="number of epochs")
    train_subparser.add_argument("-nr", "--number-repeats", type=int, help="number of repeats")
    train_subparser.add_argument("-pt", "--problem-type", help="problem type (regression or a type of classification)")
    train_subparser.add_argument("-ns", "--train-size", type=float, help="train size")
    train_subparser.add_argument("-vs", "--val-size", type=float, help="val size")
    train_subparser.add_argument("-ts", "--test-size", type=float, help="test size")
    train_subparser.add_argument("-s", "--sampler", help="choice of sampler, i.e. random, kmeans, etc.")
    train_subparser.add_argument("-rs", "--random-seed", type=int, help="random seed for sampling and pytorch seed")
    train_subparser.add_argument("-pc", "--patience", type=int, help="number of epochs to wait before early stopping")

    predict_subparser = subparsers.add_parser("predict")
    predict_subparser.add_argument("checkpoints_dir", help="directory of checkpoint file(s) for predictions")
    input_group = predict_subparser.add_mutually_exclusive_group()
    input_group.add_argument("-s", "--smiles", nargs="+", help="SMILES string for prediction")
    input_group.add_argument("-i", "--input-file", help="file containing SMILES strings")
    predict_subparser.add_argument("-o", "--output", required=False, help="output file for predictions (defaults to stdout)")

    shap_subparser = subparsers.add_parser("shap")
    shap_subparser.add_argument("checkpoints_dir", help="directory of checkpoint file(s) for SHAP analysis")
    shap_subparser.add_argument("input_file", help="csv of SMILES and targets used during training")
    shap_subparser.add_argument(
        "-it",
        "--importance-threshold",
        default=0.75,
        type=float,
        help="[0-1] hide features below (most important feature) x this",
    )

    # no args provided - print the help text
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()
    if args.version:
        print(version("fastprop"))
        exit(0)

    # cast to a dict
    args = vars(args)
    subcommand = args.pop("subcommand")
    args.pop("version")
    if subcommand == "train":
        training_default = dict(DEFAULT_TRAINING_CONFIG)
        # exit with help if no args given
        if not sum(map(lambda i: i is not None, args.values())):
            train_subparser.print_help()
            exit(0)
        optim_requested = args.pop("optimize")
        if args["config_file"] is not None:
            if sum(map(lambda i: i is not None, args.values())) > 1:
                raise parser.error("Cannot specify config_file with other command line arguments (except --optimize).")
            with open(args["config_file"], "r") as f:
                cfg = yaml.safe_load(f)
                cfg["target_columns"] = cfg["target_columns"].split(" ")
                training_default.update(cfg)
        else:
            training_default.update({k: v for k, v in args.items() if v is not None})

        optim_requested = training_default.pop("optimize") or optim_requested
        logger.info(f"Training Parameters:\n{yaml.dump(training_default, sort_keys=False)}")
        # validate this dictionary, i.e. layer counts are positive, etc.
        # cannot specify both precomputed and descriptors or enable/cache
        validate_config(training_default)
        if optim_requested:
            training_default.pop("fnn_layers")
            training_default.pop("hidden_size")
            if any((args.get("fnn_layers") is not None, args.get("hidden_size") is not None)):
                logger.warning("Hidden Size/FNN Layers specified with optimize and are ignored.")
            hopt_fastprop(**training_default)
        else:
            train_fastprop(**training_default)
    elif subcommand == "shap":
        shap_fastprop(**args)
    elif subcommand == "predict":
        if args["smiles"] is None and args["input_file"] is None:
            raise parser.error("One of -i/--input-file or -s/--smiles must be provided.")
        logger.info(f"Predict Parameters:\n {yaml.dump(args, sort_keys=False)}")
        predict_fastprop(**args)
    else:
        logger.critical(f"Unrecognized subcommand '{subcommand}', printing help and exiting.")
        parser.print_help()
        sys.exit(0)
    logger.info("If you use fastprop in published work, please cite: ...WIP...")
    logger.info("Total elapsed time: " + str(datetime.timedelta(seconds=perf_counter() - cli_start)))



import os
import pickle

import matplotlib.pyplot as plt
import numpy as np
import torch
import yaml
from tqdm import tqdm

from fastprop.defaults import init_logger
from fastprop.fastprop_core import fastprop
from fastprop.utils import calculate_mordred_desciptors, load_from_csv
from fastprop.utils.select_descriptors import mordred_descriptors_from_strings

shap = None
try:
    import shap
except ImportError as ie:
    shape_error = ie


logger = init_logger(__name__)


def shap_fastprop(checkpoints_dir, input_file, importance_threshold=0.75):
    if shap is None:
        raise RuntimeError("Unable to import shap dependencies, please install fastprop[shap]. Original error: " + str(shape_error))
    # get the configuration file with some run metadata
    checkpoint_dir_contents = os.listdir(checkpoints_dir)
    with open(os.path.join(checkpoints_dir, "fastprop_config.yml")) as file:
        config_dict = yaml.safe_load(file)

    # calculate the features in the same way the model did when training,
    # downsampling to the shap suggested size where required
    # (also see https://github.com/shap/shap/issues/955 for an explanation of
    # Why the thresholds are the way they are)
    targets, mols, smiles = load_from_csv(input_file, config_dict["smiles"], config_dict["targets"])
    num_mols = len(mols)
    threshold_1 = 200
    threshold_2 = 2000
    if num_mols > threshold_1:
        rng = np.random.default_rng(42)
        target = threshold_1 if num_mols < threshold_2 else threshold_2
        logger.info(f"Randomly downsampling to {target} molecules.")
        mols = rng.choice(mols, size=target, replace=False)
    descs = calculate_mordred_desciptors(
        mordred_descriptors_from_strings(config_dict["descriptors"]),
        mols,
        n_procs=0,  # ignored for "strategy='low-memory'"
        strategy="low-memory",
    )

    # load each of the models in the group
    all_models = []
    for checkpoint in checkpoint_dir_contents:
        if not checkpoint.endswith(".ckpt"):
            continue
        model = fastprop.load_from_checkpoint(
            os.path.join(checkpoints_dir, checkpoint),
            cleaned_data=None,
            targets=None,
            smiles=None,
        )
        try:
            with open(os.path.join(checkpoints_dir, f"repetition_{checkpoint.split('-')[1]}_scalers.yml")) as file:
                scalers = yaml.safe_load(file)
        except FileNotFoundError:
            logger.error(f"checkpoints directory is missing 'repetition_{checkpoint.split('-')[1]}_scalers.yml'. Re-execute training.")
        model.target_scaler = pickle.loads(scalers["target_scaler"])
        model.mean_imputer = pickle.loads(scalers["mean_imputer"])
        model.feature_scaler = pickle.loads(scalers["feature_scaler"])
        all_models.append(model)

    # we will use half of the data for 'integrating', and the other half for getting shap values
    halfway_idx = descs.shape[0] // 2

    # shap terminology explanation:
    # background: 100 to 1000 random samples from the training data
    # X: 100+ other samples
    # run shap on each of these models, then average the results
    per_model_shap = []
    for model in tqdm(all_models, desc="Calculating SHAP values for each model"):
        # now scale and send the data to whatever device lightning put the model on
        X = torch.tensor(model.feature_scaler.transform(model.mean_imputer.transform(descs)), dtype=torch.float32).to(model.device)
        e = shap.DeepExplainer(model, X[:halfway_idx])
        model_shap_values = e.shap_values(X[halfway_idx:])
        # returns a list for multi-target problems, cast for uniformity
        if not isinstance(model_shap_values, list):
            model_shap_values = [model_shap_values]
        per_model_shap.append(model_shap_values)
    all_shap_values = np.mean(np.array(per_model_shap), axis=0)
    # TODO: write these to a file, sorted by highest absolute SHAP on first target.

    # for each target, create a plot of the most important features
    for i, target_name in enumerate(config_dict["targets"]):
        shap_values = all_shap_values[i]
        # include features until the shap value is half the highest, aka half as important
        avg_shaps = np.mean(np.abs(shap_values), axis=0)
        avg_shaps, names, keep_idxs = zip(*sorted(zip(avg_shaps, config_dict["descriptors"], list(range(len(avg_shaps)))), reverse=True))
        include_idx = 0
        for val in avg_shaps[1:]:
            if val > avg_shaps[0] * importance_threshold:
                include_idx += 1
            else:
                break
        explanation = shap.Explanation(values=shap_values[:, keep_idxs[:include_idx]], feature_names=names[0:include_idx])
        plt.cla()
        axes = shap.plots.beeswarm(explanation, max_display=include_idx + 1, color_bar=False, color="shap_red", show=False)
        out_fname = target_name + "_feature_importance_beeswarm.png"
        if os.path.exists(out_fname):
            logger.warning(f"Output file exists! {__name__} will overwrite '{out_fname}'.")
        plt.savefig(out_fname, pad_inches=0.5, bbox_inches="tight")
    logger.info("Visit the mordred-community docs to lookup descriptors: https://jacksonburns.github.io/mordred-community/descriptors.html")



from .defaults import DEFAULT_TRAINING_CONFIG
from .hopt import hopt_fastprop
from .predict import predict_fastprop
from .preprocessing import preprocess
from .shap import shap_fastprop
from .train import train_fastprop

__all__ = ["DEFAULT_TRAINING_CONFIG", "hopt_fastprop", "predict_fastprop", "preprocess", "shap_fastprop", "train_fastprop"]



"""
fastprop - molecular property prediction with simple descriptors

This file contains the core data handling and model definition functionality
for the fastprop framework.
"""

import datetime
import os
import pickle
import warnings
from time import perf_counter
from types import SimpleNamespace
from typing import OrderedDict

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import yaml
from astartes import train_val_test_split
from astartes.molecules import train_val_test_split_molecules
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.loggers import TensorBoardLogger
from scipy.stats import ttest_ind
from sklearn.impute import SimpleImputer
from sklearn.metrics import mean_absolute_percentage_error as mape
from sklearn.metrics import mean_squared_error as l2_error
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from torch.utils.data import Dataset as TorchDataset
from torchmetrics.functional.classification import (
    auroc,
    binary_accuracy,
    f1_score,
    multiclass_accuracy,
    multiclass_auroc,
    multilabel_auroc,
)
from torchmetrics.functional.regression import r2_score

from fastprop.defaults import init_logger

logger = init_logger(__name__)

warnings.filterwarnings(action="ignore", message=".*does not have many workers which may be a bottleneck.*")

# device configuration
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
NUM_WORKERS = 1
# controls level of output
NUM_VALIDATION_CHECKS = 20


def _mock_inverse_transform(x):
    return x


class ArbitraryDataset(TorchDataset):
    """Basic PyTorch dataset class.

    Args:
        TorchDataset (pytorch.Dataset): PyTorch's dataset class.
    """

    def __init__(self, data, targets):
        self.data = data
        self.length = len(targets)
        self.targets = targets

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        return self.data[index], self.targets[index]


class fastprop(pl.LightningModule):
    """Core fastprop LightningModule

    Args:
        pl (pl.LightningModule): Parent Lightning Module class.
    """

    def __init__(
        self,
        num_epochs,
        input_size,
        hidden_size,
        readout_size,
        learning_rate,
        fnn_layers,
        problem_type,
        cleaned_data,
        targets,
        target_names,
        batch_size,
        random_seed,
        train_size,
        val_size,
        test_size,
        sampler,
        smiles,
        verbose=True,
    ):
        """Core fastprop model.

        Args:
            feature_scaler (sklearn scaler): Scaler used on feature variables, used for reporting metrics in human-scale.
            target_scaler (sklearn scaler): Scaler used on target variables, used for reporting metrics in human-scale.
            num_epochs (int): Maximum allowed number of training epochs.
            hidden_size (int): Number of neurons in the hidden layers.
            learning_rate (float): Learning rate.
            fnn_layers (int): Number of layers in the FNN.
            problem_type (str): Problem type, i.e. regression, multiclass, multilabel, or binary.
            verbose (bool, optional): Reduces some logging if true. Defaults to False.
            num_classes (int, optional): Number of classes for multiclass classification. Defaults to None.
            cleaned_data (numpy.ndarray or torch.Tensor): Descriptors, already subjected to preprocessing (dropping operations)
            targets (numpy.ndarray or torch.Tensor): Scaled targets in the same order as the descriptors.
            batch_size (int): Samples/per batch - for small feature sets like in fastprop, set as high as possible.
            random_seed (int): Seed for RNG.
            train_size (float): Fraction of data for training.
            val_size (float): Fraction of data for validation.
            test_size (float): Fraction of data for test.
            sampler (str): Type of sampler to use, see astartes for a list of implemented samplers.
            smiles (list[str], optional): SMILES strings corresponding to the molecules for use in some samplers. Defaults to None.
        """
        super().__init__()
        # used for data preparation and training
        self.random_seed = random_seed
        self.problem_type = problem_type
        self.num_classes = readout_size if self.problem_type == "multiclass" else None

        # used only for training
        self.training_metric, self.overfitting_metric = fastprop.get_metrics(problem_type)
        self.verbose = verbose
        self.num_epochs = num_epochs
        self.learning_rate = learning_rate
        # fully-connected nn
        layers = OrderedDict(
            [
                ("lin1", torch.nn.Linear(input_size, hidden_size)),
                ("act1", torch.nn.ReLU()),
            ]
        )
        for i in range(fnn_layers - 1):
            layers[f"lin{i+2}"] = torch.nn.Linear(hidden_size, hidden_size)
            layers[f"act{i+2}"] = torch.nn.ReLU()
        self.fnn = torch.nn.Sequential(layers)
        self.readout = torch.nn.Linear(hidden_size, readout_size)

        # used only for data preparation
        self.data = cleaned_data
        self.targets = targets
        self.batch_size = batch_size
        self.sampler = sampler
        self.smiles = smiles
        self.train_size, self.val_size, self.test_size = train_size, val_size, test_size
        self.train_idxs, self.val_idxs, self.test_idxs = None, None, None
        # either using folds or random sampling, both of these will be incremented by the training loop
        self.fold_number = 0
        self.target_names = target_names

        # add derived attributes to the hparams attribute manually so that they are saved with checkpoints
        self.hparams["num_classes"] = self.num_classes
        self.save_hyperparameters(ignore=("cleaned_data", "targets", "smiles"))

    def _split(self):
        logger.info(f"Sampling dataset with {self.sampler} sampler.")
        split_kwargs = dict(
            train_size=self.train_size,
            val_size=self.val_size,
            test_size=self.test_size,
            sampler=self.sampler,
            random_state=self.random_seed,
            return_indices=True,
        )
        if self.sampler != "scaffold":
            (
                *_,
                self.train_idxs,
                self.val_idxs,
                self.test_idxs,
            ) = train_val_test_split(
                self.data,
                # flatten 1D targets
                self.targets.flatten() if self.targets.shape[1] == 1 else self.targets,
                **split_kwargs,
            )
        else:
            (
                *_,
                self.train_idxs,
                self.val_idxs,
                self.test_idxs,
            ) = train_val_test_split_molecules(self.smiles, **split_kwargs)

    def setup(self, stage=None):
        if stage == "fit":
            self._split()
            logger.info("Imputing and rescaling input features.")
            # it is possible that the randomly chosen training set can have all entries missing a descriptor
            # even if other members of the dataset are valued! They will be zeroed out
            self.mean_imputer = SimpleImputer(missing_values=np.nan, strategy="mean", keep_empty_features=True)
            self.mean_imputer.fit(self.data[self.train_idxs, :])
            self.data = self.mean_imputer.transform(self.data)

            # scale each column to unit variance
            self.feature_scaler = StandardScaler()
            self.feature_scaler.fit(self.data[self.train_idxs, :])
            self.data = self.feature_scaler.transform(self.data)

            # mock the scaler object for classification tasks
            self.target_scaler = SimpleNamespace(n_features_in_=self.targets.shape[1], inverse_transform=_mock_inverse_transform)
            if self.problem_type == "regression":
                logger.info("Rescaling targets.")
                self.target_scaler = StandardScaler()
                self.target_scaler.fit(self.targets[self.train_idxs, :])
                self.targets = self.target_scaler.transform(self.targets)
            elif self.problem_type == "multiclass":
                logger.info("One-hot encoding target values.")
                self.target_scaler = OneHotEncoder(sparse_output=False)
                self.targets = self.target_scaler.fit_transform(self.targets)

            # finally, cast everything as needed
            self.data = torch.tensor(self.data, dtype=torch.float32)
            self.targets = torch.tensor(self.targets, dtype=torch.float32)

    def _init_dataloader(self, shuffle, idxs):
        return torch.utils.data.DataLoader(
            ArbitraryDataset(
                [self.data[i] for i in idxs],
                [self.targets[i] for i in idxs],
            ),
            batch_size=self.batch_size,
            shuffle=shuffle,
            num_workers=NUM_WORKERS,
            persistent_workers=True,
        )

    def train_dataloader(self):
        return self._init_dataloader(True, self.train_idxs)

    def val_dataloader(self):
        return self._init_dataloader(False, self.val_idxs)

    def test_dataloader(self):
        return self._init_dataloader(False, self.test_idxs)

    def get_metrics(problem_type):
        if problem_type == "regression":
            return "mse", "rmse"
        elif problem_type == "multilabel":
            return "bce", "auroc"
        elif problem_type == "multiclass":
            return "kldiv", "auroc"
        elif problem_type == "binary":
            return "bce", "accuracy"
        else:
            raise RuntimeError(f"Unsupported problem type '{problem_type}'!")

    def forward(self, x):
        x = self.fnn.forward(x)
        x = self.readout(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return {"optimizer": optimizer}

    def training_step(self, batch, batch_idx):
        loss = self._machine_loss(batch, reduction="mean")
        self.log(f"train_{self.training_metric}_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, _, y, y_hat = self._machine_loss(batch, return_all=True)
        self.log(f"validation_{self.training_metric}_loss", loss)
        self._human_loss(y_hat.detach().cpu(), y.detach().cpu(), "validation")
        return loss

    def test_step(self, batch, batch_idx):
        loss, _, y, y_hat = self._machine_loss(batch, return_all=True)
        self.log(f"test_{self.training_metric}_loss", loss)
        self._human_loss(y_hat.detach().cpu(), y.detach().cpu(), "test")
        return loss

    def predict_step(self, X):
        X = self.mean_imputer.transform(X)
        X = self.feature_scaler.transform(X)
        X = torch.tensor(X, dtype=torch.float32, device=self.device)
        with torch.inference_mode():
            logits = self.forward(X)
        if self.problem_type == "regression":
            return self.target_scaler.inverse_transform(logits.detach().cpu())
        elif self.problem_type in {"multilabel", "binary"}:
            return torch.sigmoid(logits).detach().cpu()
        elif self.problem_type == "multiclass":
            return torch.nn.functional.softmax(logits, dim=1).detach().cpu()

    def _machine_loss(self, batch, reduction="mean", return_all=False):
        x, y = batch
        y_hat = self.forward(x)
        if self.problem_type == "regression":
            loss = torch.nn.functional.mse_loss(y_hat, y, reduction=reduction)
        elif self.problem_type in {"multilabel", "binary"}:
            loss = torch.nn.functional.binary_cross_entropy_with_logits(y_hat, y, reduction=reduction)
        else:
            y_pred = torch.softmax(y_hat, dim=1)
            loss = torch.nn.functional.kl_div(y_pred.log(), y, reduction="batchmean")
        if not return_all:
            return loss
        else:
            return loss, x, y, y_hat

    def _human_loss(self, pred, real, name):
        n_tasks = len(self.target_names)
        # outputs the performance in more human-interpretable manner
        # expensive so we don't do this during training
        if self.problem_type == "regression":
            rescaled_pred = self.target_scaler.inverse_transform(pred)
            rescaled_truth = self.target_scaler.inverse_transform(real)
            # report r2
            if n_tasks == 1:
                r2 = r2_score(torch.tensor(rescaled_pred), torch.tensor(rescaled_truth))
                self.log(f"{name}_r2", r2)

            # mean absolute percentage error
            per_task_mape = np.abs(mape(rescaled_truth, rescaled_pred, multioutput="raw_values"))
            if n_tasks == 1:
                self.log(f"{name}_mape", np.mean(per_task_mape))
            else:
                self.log(f"{name}_mean_mape", np.mean(per_task_mape))
                for target, value in zip(self.target_names, per_task_mape):
                    self.log(f"{name}_mape_output_{target}", value)
            # same, but weighted
            # sklearn asks for an array of weights, but it actually just passes through to np.average which
            # accepts weights of the same shape as the inputs
            per_task_wmape = np.abs(mape(rescaled_truth, rescaled_pred, multioutput="raw_values", sample_weight=rescaled_truth))
            if n_tasks == 1:
                self.log(f"{name}_wmape", np.mean(per_task_wmape))
            else:
                self.log(f"{name}_mean_wmape", np.mean(per_task_wmape))
                for target, value in zip(self.target_names, per_task_wmape):
                    self.log(f"{name}_wmape_output_{target}", value)

            # mean absolute error
            all_loss = torch.nn.functional.l1_loss(torch.tensor(rescaled_pred), torch.tensor(rescaled_truth), reduction="none")
            per_task_loss = all_loss.numpy().mean(axis=0)
            if n_tasks == 1:
                self.log(f"{name}_l1", np.mean(per_task_loss))
                self.log(f"{name}_mdae", np.median(all_loss))
            else:
                self.log(f"{name}_l1_avg", np.mean(per_task_loss))
                for target, value in zip(self.target_names, per_task_loss):
                    self.log(f"{name}_l1_output_{target}", value)

            # rmse
            per_task_loss = l2_error(rescaled_truth, rescaled_pred, multioutput="raw_values", squared=False)
            if n_tasks == 1:
                self.log(f"{name}_rmse", np.mean(per_task_loss))
            else:
                self.log(f"{name}_rmse_avg", np.mean(per_task_loss))
                for target, value in zip(self.target_names, per_task_loss):
                    self.log(f"{name}_rmse_output_{target}", value)
        else:
            if self.problem_type == "multilabel":
                pred_prob = torch.sigmoid(pred)
                auroc_score = multilabel_auroc(pred_prob, real.int(), n_tasks, "macro")
                self.log(f"{name}_auroc", auroc_score)
            elif self.problem_type == "multiclass":
                pred_prob = torch.nn.functional.softmax(pred, dim=1)
                real_class = torch.tensor(self.target_scaler.inverse_transform(real.int()), dtype=torch.long).squeeze()
                score = multiclass_auroc(pred, real_class, self.num_classes)
                self.log(f"{name}_auroc", score)
                score = multiclass_accuracy(pred, real_class, self.num_classes)
                self.log(f"{name}_accuracy", score)
            elif self.problem_type == "binary":
                pred_prob = torch.sigmoid(pred)
                acc = binary_accuracy(pred_prob, real)
                self.log(f"{name}_accuracy", acc)
                f1 = f1_score(pred_prob, real, "binary")
                self.log(f"{name}_f1", f1)
                auroc_score = auroc(pred_prob, real.int(), task="binary")
                self.log(f"{name}_auroc", auroc_score)

    def on_train_epoch_end(self) -> None:
        # print progress every NUM_VALIDATION_CHECKS epochs, unless the total number of epochs is tiny (<NUM_VALIDATION_CHECKS)
        if self.num_epochs > NUM_VALIDATION_CHECKS and (self.trainer.current_epoch + 1) % (self.num_epochs // NUM_VALIDATION_CHECKS) == 0:
            if self.verbose:
                logger.info(f"Epoch [{self.trainer.current_epoch + 1}/{self.num_epochs}]")


def train_and_test(
    outdir,
    lightning_module,
    patience=5,
    verbose=True,
    no_logs=False,
    enable_checkpoints=True,
):
    """Run the lightning trainer loop on a given model.

    Args:
        outdir (str): Output directory for log files.
        datamodule (ArbitraryDataModule): Lightning-style datamodule.
        model (LightingModule): fastprop model architecture itself.
        patience (int, optional): Maximum number of epochs to wait before stopping early. Defaults to 5.
        verbose (bool, optional): Set to false for less output. Defaults to True.
        no_logs (bool, optional): Set to true to disable logs. Defaults to False.
        enable_checkpoints (bool, optional): Set to false to disable checkpoint writing. Defaults to True.

    Returns:
        list[dict{metric: score}]: Output of lightning model.test and model.validate
    """
    if not no_logs:
        try:
            repetition_number = len(os.listdir(os.path.join(outdir, "tensorboard_logs"))) + 1
        except FileNotFoundError:
            repetition_number = 1
        tensorboard_logger = TensorBoardLogger(outdir, name="tensorboard_logs", version=f"repetition_{repetition_number}", default_hp_metric=False)

    callbacks = [
        EarlyStopping(
            monitor=f"validation_{lightning_module.training_metric}_loss",
            mode="min",
            verbose=False,
            patience=patience,
        )
    ]
    if enable_checkpoints:
        callbacks.append(
            ModelCheckpoint(
                monitor=f"validation_{lightning_module.training_metric}_loss",
                dirpath=os.path.join(outdir, "checkpoints"),
                filename=f"repetition-{repetition_number}" + "-{epoch:02d}-{val_loss:.2f}",
                save_top_k=1,
                mode="min",
            )
        )

    trainer = pl.Trainer(
        max_epochs=lightning_module.num_epochs,
        accelerator=DEVICE,
        enable_progress_bar=False,
        enable_model_summary=verbose,
        logger=False if no_logs else tensorboard_logger,
        log_every_n_steps=0 if no_logs else 1,
        enable_checkpointing=enable_checkpoints,
        check_val_every_n_epoch=1,
        callbacks=callbacks,
        deterministic=True,
    )

    t1_start = perf_counter()
    trainer.fit(lightning_module)
    t1_stop = perf_counter()
    if verbose:
        logger.info("Elapsed time during training: " + str(datetime.timedelta(seconds=t1_stop - t1_start)))
    validation_results = trainer.validate(lightning_module, verbose=False)
    test_results = trainer.test(lightning_module, verbose=False)
    if verbose:
        validation_results_df = pd.DataFrame.from_records(validation_results, index=("value",))
        logger.info("Displaying validation results for repetition %d:\n%s", repetition_number, validation_results_df.transpose().to_string())
        test_results_df = pd.DataFrame.from_records(test_results, index=("value",))
        logger.info("Displaying validation results for repetition %d:\n%s", repetition_number, test_results_df.transpose().to_string())
    return test_results, validation_results


def _training_loop(
    number_repeats,
    number_epochs,
    input_size,
    hidden_size,
    readout_size,
    learning_rate,
    fnn_layers,
    output_directory,
    patience,
    problem_type,
    train_size,
    val_size,
    test_size,
    sampler,
    smiles,
    cleaned_data,
    targets,
    target_names,
    batch_size,
    random_seed,
    hopt=False,
):
    """Outer training loop to perform repeats.

    Args:
        number_repeats (int): Number of repetitions.
        number_features (int): Number of features in the input layer.
        target_scaler (sklearn scaler): Scaler used on target variables, used for reporting metrics in human-scale.
        number_epochs (int): Maximum allowed number of training epochs.
        hidden_size (int): Number of neurons in the hidden layers.
        learning_rate (float): Learning rate.
        fnn_layers (int): Number of layers in the FNN.
        problem_type (str): Problem type, i.e. regression, multiclass, multilabel, or binary.
        num_classes (int, optional): Number of classes for multiclass classification. Defaults to None.
        output_directory (str): Output directory for log files.
        datamodule (pl.DataModule): Basic data module.
        patience (int): Maximum number of epochs to wait before stopping training early.
        hopt (bool, optional): Set to true when running hyperparameter optimization to turn off logs, logging, etc. Defaults to False.

    Returns:
        list[dict{metric: score}]: Output of lightning model.test and model.validate, one pair per repetition.
    """
    if not hopt:
        logger.info(f"Run 'tensorboard --logdir {os.path.join(output_directory, 'tensorboard_logs')}' to track training progress.")
    # hopt disables some printing from fastprop, as well as the training loop, disables logging, and disables writing checkpoints
    all_test_results, all_validation_results = [], []
    for i in range(number_repeats):
        # reinitialize model
        lightning_module = fastprop(
            number_epochs,
            input_size,
            hidden_size,
            readout_size,
            learning_rate,
            fnn_layers,
            problem_type,
            cleaned_data,
            targets,
            target_names,
            batch_size,
            random_seed,
            train_size,
            val_size,
            test_size,
            sampler,
            smiles,
            verbose=not hopt,
        )
        logger.info(f"Training model {i+1} of {number_repeats}")
        test_results, validation_results = train_and_test(
            output_directory,
            lightning_module,
            patience=patience,
            verbose=not hopt,
            no_logs=hopt,
            enable_checkpoints=not hopt,
        )
        all_test_results.append(test_results[0])
        all_validation_results.append(validation_results[0])
        # save the scalers for later use
        if not hopt:
            with open(os.path.join(output_directory, "checkpoints", f"repetition_{i+1}_scalers.yml"), "w") as file:
                file.write(
                    yaml.dump(
                        dict(
                            target_scaler=pickle.dumps(lightning_module.target_scaler),
                            mean_imputer=pickle.dumps(lightning_module.mean_imputer),
                            feature_scaler=pickle.dumps(lightning_module.feature_scaler),
                        ),
                        sort_keys=False,
                    )
                )

        # resample for repeat trials
        if i + 1 < number_repeats:
            random_seed += 1
            # ensure that the model is re-initialized on the next iteration
            del lightning_module

    validation_results_df = pd.DataFrame.from_records(all_validation_results)
    logger.info("Displaying validation results:\n%s", validation_results_df.describe().transpose().to_string())
    test_results_df = pd.DataFrame.from_records(all_test_results)
    logger.info("Displaying testing results:\n%s", test_results_df.describe().transpose().to_string())
    if number_repeats > 1:
        # regression results are reported per-task, so we might need to average a differently-named
        # column
        n_tasks = len(lightning_module.target_names)
        addendum = "_avg" if (problem_type == "regression" and n_tasks > 1) else ""
        ttest_result = ttest_ind(
            test_results_df[f"test_{lightning_module.overfitting_metric}" + addendum].to_numpy(),
            validation_results_df[f"validation_{lightning_module.overfitting_metric}" + addendum].to_numpy(),
        )
        if (p := ttest_result.pvalue) < 0.05:
            logger.warn(
                "Detected possible over/underfitting! 2-sided T-test between validation and testing"
                f" {lightning_module.overfitting_metric} yielded {p=:.3f}<0.05. Consider changing patience."
            )
        else:
            logger.info(f"2-sided T-test between validation and testing {lightning_module.overfitting_metric} yielded p value of {p=:.3f}>0.05.")
    else:
        logger.info("fastprop is unable to generate statistics to check for overfitting, consider increasing 'num_repeats' to at least 2.")
    return test_results_df, validation_results_df



from importlib.metadata import version

from mordred import descriptors, get_descriptors_in_module
from packaging.version import Version


def mordred_descriptors_from_strings(string_list, include_3d=False):
    """
    returns mordred descriptor classes based on their string name
    if string_list is empty, returns all
    include_3d will include or exclude descriptors that require 3d info
    """
    out = []
    string_set = set(string_list)
    if (input_length := len(string_list)) != (set_length := len(string_set)):
        raise RuntimeError(f"Input contains {input_length - set_length} duplicate entires.")
    __version__ = Version(version("mordredcommunity"))
    for mdl in descriptors.all:
        for Desc in get_descriptors_in_module(mdl, submodule=False):
            for desc in Desc.preset(__version__):
                if not string_list or desc.__str__() in string_list:
                    if desc.require_3D and not include_3d:
                        continue
                    out.append(desc)
    return out



import os
import shutil
import unittest
from pathlib import Path

import yaml

from fastprop import train_fastprop


class Test_fastprop(unittest.TestCase):
    """
    Run a shorter benchmark as a functional test, check for regression.
    """

    @classmethod
    def setUpClass(cls):
        cls.benchmark_dir = os.path.join(Path(os.path.dirname(__file__)).parent, "benchmarks")
        cls.config_file = os.path.join(cls.benchmark_dir, "pah", "pah.yml")
        cls.temp_dirname = os.path.join(os.path.dirname(__file__), "temp")

    def test_pah(self):
        """Run the PAH benchmark."""
        with open(self.config_file, "r") as f:
            fastprop_args = yaml.safe_load(f)
            fastprop_args["target_columns"] = fastprop_args["target_columns"].split(" ")
            fastprop_args["output_directory"] = self.temp_dirname
            fastprop_args["input_file"] = os.path.join(self.benchmark_dir, "pah", "arockiaraj_pah_data.csv")
            res, _ = train_fastprop(**fastprop_args)
            assert res.describe().loc["mean", "test_r2"] > 0.95

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.temp_dirname, ignore_errors=True)


if __name__ == "__main__":
    unittest.main()



import logging
import os
from types import MappingProxyType

_logging_format = dict(
    format="[%(asctime)s %(name)s] %(levelname)s: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    level=logging.INFO,
)


def _init_loggers(outpath):
    logging.basicConfig(
        **_logging_format,
        handlers=[logging.FileHandler(os.path.join(outpath, "fastprop_log.txt")), logging.StreamHandler()],
    )


def init_logger(module_name):
    logger = logging.getLogger(module_name)
    logger.setLevel(logging.DEBUG)
    return logger


# immutable default settings
DEFAULT_TRAINING_CONFIG = MappingProxyType(
    dict(
        descriptors="optimized",
        enable_cache=True,
        precomputed=None,
        zero_variance_drop=False,
        colinear_drop=False,
        fnn_layers=2,
        learning_rate=0.0001,
        batch_size=2048,
        number_epochs=100,
        number_repeats=1,
        problem_type="regression",
        checkpoint=None,
        train_size=0.8,
        val_size=0.1,
        test_size=0.1,
        sampler="random",
        random_seed=0,
        hidden_size=1800,
        optimize=False,
        patience=5,
    )
)



"""
hopt.py

Implements parallel hyperparameter optimization for fastprop using ray[tune] and Optuna.
"""

import logging
import os

import numpy as np
import yaml

from fastprop.defaults import init_logger
from fastprop.fastprop_core import _training_loop, fastprop
from fastprop.preprocessing import preprocess
from fastprop.utils import _get_descs, load_from_csv

tune, OptunaSearch = None, None
try:
    import ray
    from ray import tune
    from ray.train.torch import enable_reproducibility
    from ray.tune.search.optuna import OptunaSearch
except ImportError as ie:
    hopt_error = ie

logger = init_logger(__name__)


CONFIG_FNAME = ".fastpropconfig"
_fpath = os.path.join(os.path.dirname(__file__), CONFIG_FNAME)
MODELS_PER_GPU, NUM_HOPT_TRIALS = 4, 64
if os.path.exists(_fpath):
    with open(_fpath) as file:
        cfg = yaml.safe_load(file)
        MODELS_PER_GPU = cfg.get("models_per_gpu", MODELS_PER_GPU)
        NUM_HOPT_TRIALS = cfg.get("num_hopt_trials", NUM_HOPT_TRIALS)


def hopt_fastprop(
    output_directory,
    input_file,
    smiles_column,
    target_columns,
    descriptors="optimized",
    enable_cache=True,
    precomputed=None,
    zero_variance_drop=True,
    colinear_drop=False,
    learning_rate=0.0001,
    batch_size=2048,
    number_epochs=1000,
    number_repeats=1,
    problem_type="regression",
    checkpoint=None,
    train_size=0.8,
    val_size=0.1,
    test_size=0.1,
    sampler="random",
    random_seed=0,
    patience=5,
    n_trials=NUM_HOPT_TRIALS,
    n_parallel=MODELS_PER_GPU,
):
    if tune is None or OptunaSearch is None:
        raise RuntimeError(
            "Unable to import hyperparameter optimization dependencies, please install fastprop[hopt].\nOriginal error: " + str(hopt_error)
        )
    logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)
    targets, mols, smiles = load_from_csv(input_file, smiles_column, target_columns)
    descs = _get_descs(precomputed, input_file, output_directory, descriptors, enable_cache, mols, as_df=True)

    logger.info("Preprocessing features")
    X = preprocess(descs, zero_variance_drop, colinear_drop).to_numpy()

    input_size = X.shape[1]
    readout_size = targets.shape[1] if problem_type != "multiclass" else (np.max(targets[:, 1]) + 1)

    X_ref = ray.put(X)
    targets_ref = ray.put(targets)
    smiles_ref = ray.put(smiles)

    _, metric = fastprop.get_metrics(problem_type)

    # driver code of optimization
    search_space = {
        "hidden_size": tune.choice(range(100, 3001, 100)),
        "fnn_layers": tune.choice(range(1, 6, 1)),
    }
    algo = OptunaSearch()
    tuner = tune.Tuner(
        tune.with_resources(
            lambda trial: objective(
                trial,
                X_ref,
                targets_ref,
                smiles_ref,
                input_size,
                readout_size,
                number_epochs,
                learning_rate,
                number_repeats,
                patience,
                problem_type,
                train_size,
                val_size,
                test_size,
                sampler,
                target_columns,
                batch_size,
                random_seed,
            ),
            # run n_parallel models at the same time (leave 20% for system)
            # don't specify cpus, and just let pl figure it out
            resources={"gpu": (1 - 0.20) / n_parallel},
        ),
        tune_config=tune.TuneConfig(
            metric=metric,
            mode="min" if problem_type == "regression" else "max",
            search_alg=algo,
            max_concurrent_trials=n_parallel,
            num_samples=n_trials,
        ),
        param_space=search_space,
    )

    results = tuner.fit()
    best = results.get_best_result().config
    logger.info(f"Best hyperparameters identified: {', '.join([key + ': ' + str(val) for key, val in best.items()])}")
    return best


def objective(
    trial,
    X_ref,
    targets_ref,
    smiles_ref,
    input_size,
    readout_size,
    number_epochs,
    learning_rate,
    number_repeats,
    patience,
    problem_type,
    train_size,
    val_size,
    test_size,
    sampler,
    target_columns,
    batch_size,
    random_seed,
) -> float:
    X = ray.get(X_ref)
    targets = ray.get(targets_ref)
    smiles = ray.get(smiles_ref)
    enable_reproducibility(random_seed)
    results_df, _ = _training_loop(
        number_repeats,
        number_epochs,
        input_size,
        trial["hidden_size"],
        readout_size,
        learning_rate,
        trial["fnn_layers"],
        None,  # output_directory ignored during hopt
        patience,
        problem_type,
        train_size,
        val_size,
        test_size,
        sampler,
        smiles,
        X,
        targets,
        target_columns,
        batch_size,
        random_seed,
        hopt=True,
    )
    _, metric = fastprop.get_metrics(problem_type)
    if readout_size == 1 or problem_type != "regression":
        return {metric: results_df.describe().at["mean", f"test_{metric}"]}
    else:
        return {metric: results_df.describe().at["mean", f"test_{metric}_avg"]}



import numpy as np
import pandas as pd
from rdkit import Chem

from fastprop.defaults import init_logger

logger = init_logger(__name__)


def load_from_csv(fpath, smiles_column, target_columns):
    # load a file of smiles and targets, returning rdkit mols and targets
    src = pd.read_csv(fpath)
    try:
        targets = src[target_columns].to_numpy()
    except KeyError as ke:
        raise RuntimeError(f"Unable to find column(s) {', '.join(target_columns)} inside file {fpath}. Check spelling and file contents.") from ke
    # keep everything 2D
    if len(target_columns) == 1:
        targets = targets.reshape(-1, 1)
    smiles: np.ndarray = src[smiles_column].to_numpy()
    logger.info("Generating RDKit molecules from SMILES.")
    rdkit_mols = np.array(list(Chem.MolFromSmiles(i) for i in smiles))

    # remove dataset entries where the molecule could not be built
    error_mols_idxs = np.where(rdkit_mols == None)[0]  # noqa: E711
    for idx in error_mols_idxs:
        logger.warn(
            f"Unable to create RDKit molecule from SMILES string {smiles[idx]} (index {idx})."
            " Both the molecule and the target will be removed from the data."
        )
    if len(error_mols_idxs) > 0:
        # specify axis=0 to avoid changing dimensions (flattening)
        logger.warn(f"Removed {len(error_mols_idxs)} entries from the dataset ({100*len(error_mols_idxs)/len(src.index):.2f}% of the data).")
        targets = np.delete(targets, error_mols_idxs, axis=0)
        rdkit_mols = np.delete(rdkit_mols, error_mols_idxs, axis=0)
        smiles = np.delete(smiles, error_mols_idxs)

    # also remove dataset entries where the target is missing
    # TODO: weight masking instead of removal?
    error_target_idxs = [idx for idx, arr in enumerate(targets) if any(np.isnan(arr))]
    for idx in error_target_idxs:
        logger.warn(
            f"Missing target value (target={targets[idx]}) for SMILES {smiles[idx]}. Both the molecule and the target will be removed from the data."
        )
    if len(error_target_idxs) > 0:
        # specify axis=0 to avoid changing dimensions (flattening)
        logger.warn(f"Removed {len(error_target_idxs)} entries from the dataset ({100*len(error_target_idxs)/len(src.index):.2f}% of the data).")
        targets = np.delete(targets, error_target_idxs, axis=0)
        rdkit_mols = np.delete(rdkit_mols, error_target_idxs, axis=0)
        smiles = np.delete(smiles, error_target_idxs)
    return targets, rdkit_mols, smiles


def load_saved_desc(fpath):
    d = pd.read_csv(fpath, low_memory=False)
    d = d.apply(pd.to_numeric, errors="coerce")
    descs = d[d.columns[1:]].to_numpy(dtype=float)
    return descs
